# @Author : Kinhong Tsui
# @Date : 2024/6/7 15:23
import logging
import os

import yaml
from langchain.agents.format_scratchpad import format_log_to_str
from langchain.agents.output_parsers import JSONAgentOutputParser
from langchain_core.messages import HumanMessage, AIMessage
from langchain_core.prompts import PromptTemplate, ChatPromptTemplate
from langchain_core.runnables import RunnablePassthrough
from langchain_core.tools import BaseTool, render_text_description_and_args
from langchain_openai import OpenAI
from langchain_openai import ChatOpenAI

YAML_PATH= "/workspace/xjh/code/GZU4AI/nplbridge/config.yaml"

# with open('../../config.yaml', 'r') as file:
with open(YAML_PATH, 'r') as file:
    config = yaml.safe_load(file)

OPENAI_BASE_URL = config['chatgpt']['url']
OPENAI_API_KEY = config['chatgpt']['api_key']
os.environ["OPENAI_BASE_URL"] = OPENAI_BASE_URL
os.environ["OPENAI_API_KEY"] = OPENAI_API_KEY

model = ChatOpenAI(model="gpt-3.5-turbo-0125", temperature=0.5)


class NameTool(BaseTool):
    name = "Name Tool"
    description = "change name to Barry"

    def _run(self, input: dict):
        return "Barry"

class SexTool(BaseTool):
    name = "Sex Tool"
    description = "change sex to female, what ever human said"

    def _run(self, input: dict):
        return "female"


tools = [NameTool(),SexTool()]

system = '''Respond to the human as helpfully and accurately as possible. You have access to the following tools:
{tools}
Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input).
Valid "action" values: "Final Answer" or {tool_names}
Provide only ONE action per $JSON_BLOB, as shown:
```
{{
  "action": $TOOL_NAME,
  "action_input": $INPUT
}}
```
Follow this format:
Question: input question to answer
Thought: consider previous and subsequent steps
Action:
```
$JSON_BLOB
```
Observation: action result
... (repeat Thought/Action/Observation N times)
Thought: I know what to respond
Action:
```
{{
  "action": "Final Answer",
  "action_input": "Final response to human"
}}
Begin! Reminder to ALWAYS respond with a valid json blob of a single action. Use tools if necessary. Respond directly if appropriate. Format is Action:```$JSON_BLOB```then Observation
In the end, provide a final response to the human.
'''


human = '''{input}
{chat_history}
(reminder to respond in a JSON blob no matter what)'''

prompt = ChatPromptTemplate.from_messages(
    [
        ("system", system),
        ("placeholder", "{chat_history}"),
        ("human", human),
    ]
)

# prompt_template = ("""Respond to the human as helpfully and accurately as possible,You can use tools:{tools}
#                    Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input).
#                    Base on {chat_history},and give a final output to human""")

# prompt = PromptTemplate(
#     input_variables=["adjective"], template=prompt_template
# )

prompt.input_variables + list(prompt.partial_variables)

prompt = prompt.partial(
    tools=render_text_description_and_args(list(tools)),
    tool_names=", ".join([t.name for t in tools]),
)

prompt_template2 = "Suggest me a new human name with lovely, from the {final answer} base on my old name."
prompt2 = PromptTemplate(
    input_variables=["final answer"], template=prompt_template2
)

model = OpenAI()

# stop = ["\n action_input"]
# model = model.bind(stop=stop)

# chain = prompt | model | JSONAgentOutputParser() | prompt2 | model
chain = ( prompt | model)
print(type(chain))

res = chain.invoke({
    "input": "what's my sex?",
    "chat_history": [
        HumanMessage(content="hi! my name is bob"),
        AIMessage(content="Hello Bob! How can I assist you today?")
    ]
})

print(f"res1: {res}")

chain2 = prompt2 | model
print(type(chain2))

res2 = chain2.invoke(res)
print(f"res2: {res2}")

#
#
# res2 = chain2.invoke(res)
# print(f"res2: {res2}")



# from nlpbridge.audio import(

# )
from nlpbridge.langchain_zhipuai.embeddings.base import ZhipuAIEmbeddings
from nlpbridge.persistent import (
    RedisDB,
)
from nlpbridge.text import (
    RedisMemory,
)
from nlpbridge.text import (
    EasyAgent,
    HappyAgent,
    ChatAgent
)
from nlpbridge.text.router import(
    RouterManager,
    Condition,
    Node,
)

__all__ = [
    "ZhipuAIEmbeddings",
    "RedisDB",
    "RedisMemory",
    "ChatAgent",
    "RouterManager",
    "Condition",
]


import logging
from functools import reduce
from typing import Any, Dict, List, Optional
from typing import Union, Sequence

from langchain.agents.format_scratchpad import format_log_to_str
from langchain.agents.output_parsers import JSONAgentOutputParser
from langchain.chains.base import Chain
from langchain_community.chat_message_histories import RedisChatMessageHistory
from langchain_core.embeddings import Embeddings
# use argparse to parse the command line arguments
from langchain_core.language_models import BaseLanguageModel
from langchain_core.memory import BaseMemory
from langchain_core.output_parsers import BaseOutputParser
from langchain_core.prompts import BasePromptTemplate, ChatPromptTemplate
from langchain_core.runnables import *
from langchain_core.runnables.history import RunnableWithMessageHistory
from langchain_core.tools import BaseTool, ToolsRenderer, render_text_description_and_args
# import vectorstore from langchain
from langchain_core.vectorstores import VectorStore

logger = logging.getLogger(__name__)


class EasyAgent(Chain):
    """
    NOTE: Following is the docstring for the Chain class:
    Chain should be used to encode a sequence of calls to components like
    models, document retrievers, other chains, etc., and provide a simple interface
    to this sequence.

    The Chain interface makes it easy to create apps that are:
        - Stateful: add Memory to any Chain to give it state,
        - Observable: pass Callbacks to a Chain to execute additional functionality,
            like logging, outside the main sequence of component calls,
        - Composable: the Chain API is flexible enough that it is easy to combine
            Chains with other components, including other Chains.

    The main methods exposed by chains are:
        - `__call__`: Chains are callable. The `__call__` method is the primary way to
            execute a Chain. This takes inputs as a dictionary and returns a
            dictionary output.
        - `run`: A convenience method that takes inputs as args/kwargs and returns the
            output as a string or object. This method can only be used for a subset of
            chains and cannot return as rich of an output as `__call__`.
    """
    memory: Optional[BaseMemory] = None
    prompt_template_manager: Optional[BasePromptTemplate] = None
    llm: Optional[BaseLanguageModel] = None
    tools: Optional[Dict[str, Any]] = {}
    vectorstore: Optional[VectorStore] = None
    embedding: Optional[Embeddings] = None
    outputParser: Optional[BaseOutputParser] = None
    runnables: Optional[List[Runnable]] = None
    input_key: str = "text"
    output_key: str = "answer"

    @classmethod
    def from_runnables(
            cls,
            runnables: List[Runnable],
            **kwargs: Any,
    ) -> "EasyAgent":
        return cls(runnables=runnables, **kwargs)

    @classmethod
    def from_llm(
            cls,
            llm: BaseLanguageModel,
            **kwargs: Any,
    ) -> "EasyAgent":
        return cls(llm=llm, **kwargs)

    @classmethod
    def from_str(
            cls,
            llm: BaseLanguageModel,
            ss: str,
    ) -> "EasyAgent":
        return cls(llm=llm)

    @property
    def input_keys(self) -> List[str]:
        return [self.input_key]

    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]

    def _call(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        if self.runnables is not None:
            return self.runnables(inputs)

        default_sequence = RunnableSequence()

        return inputs

    def set_chain_structure(self, runnables: List[Runnable]):
        """
        user can define the chain structure by passing a Runnable object.
        e.g.: RunnableSequence, RunnableParallel, RunnableLambda, etc.

        if runnables is None, the chain will be a simple chain with a single Runnable or default RunnableSequence
        """
        self.runnables = runnables

    def prep_inputs(self, inputs: Union[Dict[str, Any], Any]) -> Dict[str, str]:
        # use memory or not:
        if self.agent_config.use_memory:
            if self.memory is None:
                raise ValueError("Memory is not set. Please set memory before using it.")
            inputs = self.super().prep_inputs(inputs)
            return inputs

        return inputs

    def run(self, inputs: Dict[str, Any], **kwargs) -> str:
        if self.runnables is not None:
            chain = reduce(lambda acc, runnable: acc | runnable, self.runnables)
            return chain.invoke(inputs, **kwargs)
        if self.llm is not None:
            return self.llm.invoke(inputs['input'], **kwargs)

        raise ValueError("No runnables or llm set for the agent")


def create_sequence_structure(runnables: List[Runnable]) -> Runnable:
    """
    create a RunnableSequence structure from the list of runnables
    """
    return RunnableSequence(steps=runnables)


def create_parrallel_structure(runnables: List[Runnable]) -> Runnable:
    """
    create a RunnableParallel structure from the list of runnables
    """
    return RunnableParallel(steps=runnables)


def create_graph_structure(runnables: Dict[str, Runnable]) -> Runnable:
    pass


def create_structured_chat_agent(
        llm: BaseLanguageModel,
        tools: Sequence[BaseTool],
        prompt: ChatPromptTemplate,
        tools_renderer: ToolsRenderer = render_text_description_and_args,
        *,
        stop_sequence: Union[bool, List[str]] = True, ) -> Runnable:
    """
    create a structured chat agent with the given runnables
    """
    missing_vars = {"tools", "tool_names", "agent_scratchpad"}.difference(
        prompt.input_variables + list(prompt.partial_variables)
    )
    if missing_vars:
        raise ValueError(f"Prompt missing required variables: {missing_vars}")

    prompt = prompt.partial(
        tools=tools_renderer(list(tools)),
        tool_names=", ".join([t.name for t in tools]),
    )

    if stop_sequence:
        stop = ["\nObservation"] if stop_sequence is True else stop_sequence
        llm_with_stop = llm.bind(stop=stop)
    else:
        llm_with_stop = llm

    agent = (
            RunnablePassthrough.assign(
                agent_scratchpad=lambda x: format_log_to_str(x["intermediate_steps"]),
            )
            | prompt
            | llm_with_stop
            | JSONAgentOutputParser()
    )

    return agent


class HappyAgent(Chain):
    chain_list: Optional[list[RunnableSequence]] = None
    chains: Optional[RunnableSequence] = None
    chains_with_history: Optional[RunnableWithMessageHistory] = None
    config: Optional[Dict[str, Any]] = None
    memory: bool = True
    history_config: Optional[Dict[str, Any]] = None
    agentId: Optional[str] = None

    input_key: str = "text"
    output_key: str = "answer"

    def __init__(self, /, **kwargs: Any):
        super().__init__(**kwargs)
        if self.chain_list is not None:
            self.set_chains_structure()
        if self.memory:
            self.set_chains_history()
        self.history_config = {"configurable": {"session_id": self.agentId}}

    @classmethod
    def from_chains(
            cls,
            chain_list: list[RunnableSequence],
            **kwargs: Any,
    ) -> "HappyAgent":
        return cls(chain_list=chain_list, **kwargs)  # chain_list, config, agentId, options: memory

    @property
    def input_keys(self) -> List[str]:
        return [self.input_key]

    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]

    def _call(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        return inputs

    def set_chains_structure(self):
        chains = reduce(lambda acc, runnable_sequence: acc | runnable_sequence, self.chain_list)
        self.chains = chains
        return True

    def set_chains_history(self):
        self.chains_with_history = RunnableWithMessageHistory(
            self.chains,
            lambda session_id: RedisChatMessageHistory(
                session_id,
                url=self.config['redis']["redis_url"],
                key_prefix="history_messages:"
            ),
            input_messages_key="inputs",
            history_messages_key="history_messages",
        )

    def run(self, inputs, memory=True, **kwargs) -> str:
        if memory:
            return self.chains_with_history.invoke({"inputs": inputs}, config=self.history_config)

        return self.chains.invoke(inputs, **kwargs)


def get_meta_chain(prompt: BasePromptTemplate, model: BaseLanguageModel, output_parser: BaseOutputParser = None):
    if output_parser is not None:
        return prompt | model | output_parser

    return prompt | model


####@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@####
from nlpbridge.text.template_manager import init_prompt_with_tools
from nlpbridge.persistent.redis import RedisDB
from datetime import datetime
import os


class ChatAgent:
    def __init__(self, yaml_path: None, agent_id, llm, tools, prompt, memory=False):
        # if yaml_path is None and os.getenv("YAML_PATH") is not None:
        #     yaml_path = os.getenv("YAML_PATH")
        # else:
        #     raise ValueError("YAML_PATH not provided.")
        self.agent_id = agent_id
        self.memory = memory
        self.config = self.load_config(yaml_path)
        self.db = RedisDB(config=self.config)
        self.agent = self.create_agent(llm, tools, prompt)
        self.history = self.setup_message_history()

    @staticmethod
    def load_config(yaml_path: None):
        from nlpbridge.config import CONFIG
        # with open(yaml_path, 'r') as file:
        #     config = yaml.safe_load(file)
        # Update the dynamic parts
        current_date = datetime.now().strftime("%Y-%m-%d")
        CONFIG.os_config.LANGCHAIN_PROJECT = current_date
        for key, value in CONFIG.dict_config['os_config'].items():
            os.environ[key] = value
        return CONFIG

    def create_agent(self, llm, tools, prompt):
        # agent = create_structured_chat_agent(llm, tools, prompt)
        # agent = create_openai_tools_agent(llm,tools, prompt)
        if tools[0] != 'None':
            prompt = init_prompt_with_tools(prompt, tools)
        agent = prompt | llm
        # return AgentExecutor(agent=agent, tools=tools, verbose=True, handle_parsing_errors=True, max_iterations=10)
        return agent

    def setup_message_history(self):
        if self.memory:
            # Setup to handle message history
            return RunnableWithMessageHistory(
                self.agent,
                lambda session_id: RedisChatMessageHistory(
                    session_id,
                    url=f"redis://{self.config.redis.user}:{self.config.redis.password}@{self.config.redis.url}:{self.config.redis.port}/0",
                    ttl=60 * 60 * 24,
                ),
                input_messages_key="input",
                history_messages_key="chat_history"
            )
        else:
            raise ValueError("Message history is disabled, ensure it is enabled before using this method.")

    async def run(self, query):
        # Use the message history system to handle queries
        if self.memory:
            return await self.history.ainvoke({"input": query},
                                       config={"configurable": {"session_id": self.agent_id}})
        else:
            return await self.agent.ainvoke({"input": query})

    def stream(self, query):
        # Use the message history system to handle queries with steam pattern
        if self.memory:
            return self.history.stream({"input": query},
                                       config={"configurable": {"session_id": self.agent_id}})
        else:
            return self.agent.stream({"input": query})

    def interactive_session(self):
        while True:
            # user_input = input("Input: ")
            user_input = "你好呀，你是谁？"
            print("Output:", self.run(user_input))



# @Author : Kinhong Tsui
# @Date : 2024/6/4 9:49

import pickle
import base64

from typing import Mapping
from langchain_core.prompts import BasePromptTemplate, ChatPromptTemplate
from langchain_core.tools import render_text_description_and_args, BaseTool

from nlpbridge.persistent.persistent_store import PersistentStore


class TemplateManager:

    def __init__(self, persistent_store: PersistentStore):
        if persistent_store is None:
            raise ValueError("Persistent store must be provided.")
        self.persistent_store = persistent_store

    def save_template(self, key,
                      template_map: Mapping[str, BasePromptTemplate] = None,
                      data_type: str = "hash"
                      ) -> bool:
        if key is None or template_map is None:
            raise ValueError("Key and template map must be provided.")
        field_value_pairs = {k: base64.b64encode(pickle.dumps(template_map[k])).decode('utf-8') for k in template_map}
        return self.persistent_store.store(key, field_value_pairs, data_type)

    def load_template(self,
                      key, fields,
                      data_type: str = "hash"
                      ) -> list[BasePromptTemplate]:
        if key is None or fields is None:
            raise ValueError("Key and fields must be provided.")
        encoded_templates = self.persistent_store.retrieve(key, fields, data_type)
        if None in encoded_templates:
            return []
        templates = [pickle.loads(base64.b64decode(template)) for template in encoded_templates]
        return templates

    def delete_template(self,
                        key,
                        fields,
                        data_type: str = "hash"
                        ) -> int:
        if key is None:
            raise ValueError("Key must be provided.")
        return self.persistent_store.delete(key, fields, data_type)


class TemplateGetter():

    def __init__(self, template_manager):
        if template_manager is None:
            raise ValueError("Template manager must be provided.")
        self.template_manager = template_manager



    def get_chat_prompt_template(self, field='default_template', tools=None) -> ChatPromptTemplate:
        if field is None:
            raise ValueError("Field must be provided.")
        chat_prompt_template = self.template_manager.load_template('chat_prompt_template', [field])[0]
        if tools is not None:
            chat_prompt_template = init_prompt_with_tools(chat_prompt_template, tools)
        return chat_prompt_template

    def define_chat_prompt_template(self, system: str = None,
                                    placeholder="{chat_history}",
                                    human: str = None) -> ChatPromptTemplate:

        if system is None or placeholder is None or human is None:
            raise ValueError("System, placeholder, and human must be provided.")

        new_chat_prompt_template = ChatPromptTemplate.from_messages(
            [
                ("system", system),
                ("placeholder", placeholder),
                ("human", human),
            ])

        return new_chat_prompt_template

def init_prompt_with_tools(prompt: BasePromptTemplate, tools: list[BaseTool]):
    prompt.input_variables + list(prompt.partial_variables)

    return prompt.partial(
        tools=render_text_description_and_args(list(tools)),
        tool_names=", ".join([t.name for t in tools]),
    )


"""
Base Tools
"""
import os
from typing import (
    List,
)

from langchain_community.utilities import SerpAPIWrapper
from langchain_community.vectorstores import Redis
from langchain_core.tools import tool

from nlpbridge.langchain_zhipuai.embeddings.base import ZhipuAIEmbeddings


@tool("web_search")
def web_search(query: str) -> str:
    """A search engine. Useful for when you need to answer questions about current events. Input should be a search query."""
    search = SerpAPIWrapper()
    return search.run(query)


@tool
def rag_google(query: str) -> List[str]:
    """Get google information from the local repository"""
    retriever = Redis(redis_url=os.getenv('REDIS_URL'),
                      index_name=os.getenv("RAG_GOOGLE_INDEX_NAME"),
                      embedding=ZhipuAIEmbeddings()
                      ).as_retriever()
    docs = retriever.invoke(input=query)
    return [doc.page_content for doc in docs]


@tool
def multiply(x: float, y: float) -> float:
    """Multiply 'x' times 'y'."""
    return x * y


@tool
def subtract(x: float, y: float) -> float:
    """subtract 'x' and 'y'."""
    return x - y


@tool
def exponentiate(x: float, y: float) -> float:
    """Raise 'x' to the 'y'."""
    return x ** y


@tool
def add(x: float, y: float) -> float:
    """Add 'x' and 'y'."""
    return x + y


tool_list = {
    'None': 'None',
    'add': add,
    'sub': subtract,
    'exp': exponentiate,
    'mul': multiply
}



# @Author : Kinhong Tsui
# @Date : 2024/6/5 10:03

import yaml
import pickle

from langchain_core.prompts import ChatPromptTemplate
from nlpbridge.persistent.redis import RedisDB
from nlpbridge.text.template_manager import TemplateManager

system = '''Respond to the human as helpfully and accurately as possible. You have access to the following tools:
{tools}
Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input).
Valid "action" values: "Final Answer" or {tool_names}
Provide only ONE action per $JSON_BLOB, as shown:
```
{{
  "action": $TOOL_NAME,
  "action_input": $INPUT
}}
```
Follow this format:
Question: input question to answer
Thought: consider previous and subsequent steps
Action:
```
$JSON_BLOB
```
Observation: action result
... (repeat Thought/Action/Observation N times)
Thought: I know what to respond
Action:
```
{{
  "action": "Final Answer",
  "action_input": "Final response to human"
}}
Begin! Reminder to ALWAYS respond with a valid json blob of a single action. Use tools if necessary. Respond directly if appropriate. Format is Action:```$JSON_BLOB```then Observation'''

human = '''{input}
{agent_scratchpad}
(reminder to respond in a JSON blob no matter what)'''

prompt = ChatPromptTemplate.from_messages(
    [
        ("system", system),
        ("placeholder", "{chat_history}"),
        ("human", human),
    ]
)


save = pickle.dumps(prompt)
print(save)
print(type(save))

print()
res = pickle.loads(save)
print(res)
print(type(res))


import sys,os
import yaml
sys.path.append(os.getcwd())
from nlpbridge.persistent.redis import RedisDB

YAML_PATH= "/Users/apple/vitalis/source_code/midplugin/config.yaml"

with open(YAML_PATH, 'r') as file:
    config = yaml.safe_load(file)

rds_store = RedisDB(config=config)

key_name = "hash_key"
def test_mhget() -> None:
    """Test mhget method."""
    fileds = ["key1", "key2"]
    rds_store.store(key_name,{"key1": "value1", "key2": "value2"})
    rds_store.store(key_name,{"key1": "value1", "key2": "value2", "key3": "value3"})
    result = rds_store.retrieve(key_name,fileds)
    assert result == ["value1", "value2"]


def test_mhset() -> None:
    """Test that multiple keys can be hash."""
    field_value_pairs = {"key1": "value1", "key2": "value2", "key3": "value3"}
    rds_store.store(key_name, field_value_pairs)
    res = rds_store.retrieve(key_name,["key1", "key2", "key3"])
    assert  res == ["value1", "value2", "value3"]


def test_mdel()-> None:
    """Test that deletion works as expected."""
    fileds = ["key1", "key2"]
    rds_store.store(key_name,{"key1": "value1", "key2": "value2"})
    rds_store.store(key_name, {"key1": "value1", "key2": "value2", "key3": "value3"})
    rds_store.delete(key_name,fileds)
    result = rds_store.retrieve(key_name,fileds)
    assert result == [None, None]

def test_hkeys():
    res = rds_store.hkeys(key_name)
    print(res)

if __name__ == "__main__":
    test_mhget()
    test_mhset()
    test_mdel()
    test_hkeys()


import contextlib
from typing import Any, AsyncIterator

from sqlalchemy import AsyncAdaptedQueuePool
from sqlalchemy.ext.asyncio import (
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.orm import DeclarativeBase

from nlpbridge.config import CONFIG


class Base(DeclarativeBase):
    # https://docs.sqlalchemy.org/en/14/orm/extensions/asyncio.html#preventing-implicit-io-when-using-asyncsession
    __mapper_args__ = {"eager_defaults": True}


# Heavily inspired by https://praciano.com.br/fastapi-and-async-sqlalchemy-20-with-pytest-done-right.html
class DatabaseSessionManager:
    def __init__(self, host: str, engine_kwargs: dict[str, Any] = {}):
        self._engine = create_async_engine(host, **engine_kwargs)
        self._sessionmaker = async_sessionmaker(autocommit=False, bind=self._engine, expire_on_commit=False)

    async def close(self):
        if self._engine is None:
            raise Exception("DatabaseSessionManager is not initialized")
        await self._engine.dispose()

        self._engine = None
        self._sessionmaker = None

    @contextlib.asynccontextmanager
    async def session(self) -> AsyncIterator[AsyncSession]:
        if self._sessionmaker is None:
            raise Exception("DatabaseSessionManager is not initialized")
        try:
            session = self._sessionmaker()
            yield session
        except Exception:
            await session.rollback()
            raise
        finally:
            await session.invalidate()


SQLALCHEMY_DATABASE_URI = f"mysql+asyncmy://{CONFIG.mysql.user}:{CONFIG.mysql.password}@{CONFIG.mysql.host}:{CONFIG.mysql.port}/{CONFIG.mysql.db}"
engine_kwargs = {"pool_size": 64, "max_overflow": 0, "pool_recycle": 3600, "pool_pre_ping": True, "echo": False,
                 "echo_pool": True, "poolclass": AsyncAdaptedQueuePool, "pool_timeout": 5}
sessionmanager = DatabaseSessionManager(SQLALCHEMY_DATABASE_URI, engine_kwargs)



from langchain_core.memory import BaseMemory
from langchain_core.runnables import Runnable
from nlpbridge.persistent.redis import RedisDB
from typing import (
    Any, Dict, Union,
    List,
    Optional,
)
import json

from nlpbridge.persistent.redis import RedisDB


class Memory(BaseMemory):
    def __init__(self, dbclient):
        self.dbclient = dbclient

    def create(self, key: str, value: Any):
        self.dbclient.set(key, value)

    def get(self, key: str) -> Any:
        return self.dbclient.get(key)
    
    def delete(self, key: str):
        self.dbclient.delete(key)


class RedisMemory(BaseMemory):
    uid: str
    agentId: str
    redis: Optional[RedisDB] | None
    message_key: str
    counter_key: str
    kwargs: Optional[Any] | None
    
    def __init__(self, uid, agentId, redis:Optional[RedisDB] | None, **kwargs: Any):
        super().__init__(uid=uid, agentId=agentId, redis=redis,message_key='',counter_key='',kwargs=kwargs)
        if uid is None or not uid:
            raise ValueError("UID cannot be None or empty")
        if agentId is None or not agentId:
            raise ValueError("Agent ID cannot be None or empty")
        if redis is None:
            raise ValueError("Redis instance cannot be None")

        self.uid = uid
        self.agentId = agentId
        self.redis = redis
        self.message_key = f"context:{self.uid}.{self.agentId}"
        self.counter_key = f"counter:{self.uid}.{self.agentId}"


    def memory_variables(self) -> List[str]:
        return self.redis.hkeys(self.message_key)

    def load_memory_variables(self, inputs: Dict[str, Any]|None) -> Dict[str, Any]:
        # raw_data = self.redis.retrieve_hash(self.message_key)
        # processed_data = {}
        # for key, value in raw_data.items():
        #     processed_data[key] = json.loads(value) if value else None
        # return processed_data
        pass

    def save_context(self, inputs: Dict[str,Any], outputs: Dict[str,str]) -> Any:
        sequence_number = self.redis.client.incr(self.counter_key)
        messages = {f"{sequence_number}": json.dumps({"inputs": list(inputs.values()), "outputs": list(outputs.values())})}
        return self.redis.store(self.message_key,messages)
    
    def clear(self):
        self.redis.delete(self.message_key,None)


# @Author : Kinhong Tsui
# @Date : 2024/6/5 16:04


import yaml

from nlpbridge.persistent.redis import RedisDB
from nlpbridge.text.template_manager import TemplateManager, TemplateGetter

with open('../../config.yaml', 'r') as file:
    config = yaml.safe_load(file)
db = RedisDB(config)

template_manager = TemplateManager(db)
template_getter = TemplateGetter(template_manager)

print(template_getter.get_chat_prompt_template("template1"))


print("=====================================")

system = "system{tools}"
human = "human{input}"

new_template = template_getter.define_chat_prompt_template(system=system, human=human)
print(new_template)


import operator
from typing import Annotated, Sequence, TypedDict
from langchain_core.messages import AIMessage, BaseMessage, HumanMessage, ToolMessage
from langchain_core.runnables import RunnableLambda
from langgraph.graph import END, StateGraph


class AgentState(TypedDict):
    messages: Annotated[Sequence[BaseMessage],operator.add]


import copy
import pydantic

__version__ = '0.0.1'


class SDKReportConfig(pydantic.BaseModel):
    sdk_version: str = __version__
    sdk_language: str = "python"


# report information
default_header = {
    "X-GZU-Sdk-Config": SDKReportConfig().model_dump_json(),
    "X-GZU-Origin": 'gzu_sdk'
}


def get_default_header():
    return copy.deepcopy(default_header)


from nlpbridge.audio._exception import (
    BadRequestException,
    ForbiddenException,
    NotFoundException,
    PreconditionFailedException,
    InternalServerErrorException,
    HTTPConnectionException,
    GZUServerException,
)

from nlpbridge.audio import *
from nlpbridge.audio.message import Message
from nlpbridge.audio.utils.logger_util import logger
from nlpbridge.audio.components.asr.asr import ASR
from nlpbridge.audio.components.tts.tts import TTS

__all__ = [
    # 日志组件
    "logger",
    # 通用入参组件
    "Message",
    # 异常定义
    "BadRequestException",
    "ForbiddenException",
    "NotFoundException",
    "PreconditionFailedException",
    "InternalServerErrorException",
    "HTTPConnectionException",
    "GZUServerException",
    # 语音识别组件
    'ASR',
    'TTS'
]



# # @Author : Kinhong Tsui
# # @Date : 2024/6/3 15:41
from typing import Sequence, Tuple

import yaml
from typing import(
    Any,
    List,
    Sequence,
    cast,
    Optional,
    Tuple,
    Mapping,
)
from typing_extensions import(
    TypeAlias,
)
from langchain_community.storage import RedisStore
from langchain_community.utilities.redis import get_client
from nlpbridge.persistent.persistent_store import PersistentStore
import logger

# _Value: TypeAlias = bytes | float | int | str
# _Key: TypeAlias = str | bytes

class RedisDB(PersistentStore,RedisStore):
    def __init__(self,config):
        self.redis_config = config.dict_config['redis']
        self.url = self.redis_config['url']
        self.port = self.redis_config['port']
        self.password = self.redis_config['password']
        self.db = self.redis_config['db']
        self.redis_url = f"redis://default:{self.password}@{self.url}:{self.port}/{self.db}"

        try:
            import redis
        except ImportError:
            raise ImportError(
                "Could not import redis python package. "
                "Please install it with `pip install redis`."
            )

        try:
            self.client = get_client(redis_url=self.redis_url,decode_responses=True)
        except redis.exceptions.ConnectionError as error:
            logger.error(error)



    '''Support hash data type storage'''
    def store(self, key, field_value_pairs: Mapping[str, str], data_type:str="hash") -> bool:
        """Set the given key-value pairs with key in hash."""
        if data_type == "hash":
            return self.client.hmset(key, field_value_pairs)
        else:
            return None

    def retrieve(self, key, fields: Sequence[str], data_type:str="hash") -> List[Optional[str]]:
        """Get the values associated with the given fields in hash."""
        if data_type == "hash":
            return cast(List[Optional[str]], self.client.hmget(key, fields))
        else:
            return None

    def update(self, key, field_value_pairs: Mapping[str, str], data_type:str="hash") -> bool:
        """Set the given key-value pairs with key in hash."""
        if data_type == "hash":
            return self.client.hmset(key, field_value_pairs)
        else:
            return None

    def delete(self, key, fields: Optional[Sequence[str]], data_type:str="hash") -> int:
        """Delete the given keys."""
        if data_type == "hash":
            if fields is None:
                return self.client.delete(key)
            return self.client.hdel(key, *fields)
        else:
            return None

    def hkeys(self, key) -> List[str]:
        """Get the values associated with the given fields in hash."""
        return self.client.hkeys(key)

    '''Support set data type storage'''

    '''Support list data type storage'''

    '''Support zset data type storage'''



import sys,os
import yaml
sys.path.append(os.getcwd())
import logging
from nlpbridge.persistent.redis import RedisDB
from typing import(
    Any,List,
)
from langchain_community.chat_message_histories import RedisChatMessageHistory
from langchain_core.runnables.history import RunnableWithMessageHistory
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.chat_history import BaseMessage
from langchain_openai import ChatOpenAI
from nlpbridge.langchain_zhipuai import ZhipuAIEmbeddings
from langchain_community.document_loaders import WebBaseLoader
from langchain_text_splitters import RecursiveCharacterTextSplitter
from langchain_core.documents import Document
from langchain_community.vectorstores import Redis
from langchain.retrievers.multi_query import MultiQueryRetriever


os.environ["ZHIPUAI_API_KEY"] = "59e910acc5d4d68b9ef544009d89198d.mUfSMG5h6xgTphfL"
YAML_PATH= "/Users/apple/vitalis/source_code/nlpbridge/config.yaml"

with open(YAML_PATH, 'r') as file:
    config = yaml.safe_load(file)

redis = config['redis']
redis["redis_url"] = f"redis://default:{redis['password']}@{redis['url']}:{redis['port']}/{redis['db']}"
os.environ["REDIS_URL"] = redis['redis_url']

url = 'https://zh.wikipedia.org/wiki/Google'
def load_blog(url:str)->List[str]:
    # Load blog post
    loader = WebBaseLoader(url)
    data = loader.load()
    text = [doc.page_content for doc in data]

    # Split
    text_splitter = RecursiveCharacterTextSplitter(chunk_size=512, chunk_overlap=0)
    splits = text_splitter.split_text(' '.join(text))
    return splits

def  redis_vector(docs:List[str])->Redis:
    return Redis(redis_url=os.getenv('REDIS_URL'),
                 index_name=config['os_config']["RAG_GOOGLE_INDEX_NAME"],
                 embedding=ZhipuAIEmbeddings(),
                 ).from_texts(texts=docs,embedding=ZhipuAIEmbeddings(),index_name=config['os_config']["RAG_GOOGLE_INDEX_NAME"])


llm =  ChatOpenAI(api_key=config['chatgpt']['api_key'],base_url=config['chatgpt']['url'],temperature=0)
retriever_from_llm = MultiQueryRetriever.from_llm(
    retriever=redis_vector(load_blog(url)).as_retriever(), llm=llm
)

logging.basicConfig()
logging.getLogger("langchain.retrievers.multi_query").setLevel(logging.ERROR)

for _ in iter(int,1):
    # user_input = input("Input: ")
    user_input = "What are some recent developments in Google search?"

    # Get unique documents using the LLM-based retriever
    # This will use the RAG (Reinforced Aggregation) model to find relevant documents from the Redis vector store
    # The retrieved documents are then returned in a list along with their scores.
    unique_docs = retriever_from_llm.invoke(user_input)
    print("\nOutput: ",unique_docs[0].page_content)



import os,sys

from langchain_community.chat_message_histories import RedisChatMessageHistory
from langchain_core.runnables.history import RunnableWithMessageHistory
from langchain_core.tools import render_text_description_and_args, BaseTool

from nlpbridge.text.agent import get_meta_chain, HappyAgent, ChatAgent
# sys.path.append(os.getcwd())
from datetime import datetime
from nlpbridge.text.tools import multiply,exponentiate,add,rag_google,subtract,web_search,tool
from langchain_openai import ChatOpenAI

import operator,functools
from typing import Annotated, Sequence, TypedDict,Any,List
from langchain_core.messages import AIMessage, BaseMessage, HumanMessage, ToolMessage
from langchain_core.runnables import RunnableLambda
from langgraph.graph import END, StateGraph 
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain.agents import AgentExecutor, create_openai_tools_agent
from langchain.output_parsers.openai_functions import JsonOutputFunctionsParser
from langchain_core.output_parsers import StrOutputParser
from langchain.agents import create_structured_chat_agent
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    ChatMessage,
    FunctionMessage,
    HumanMessage,
    SystemMessage
)

os.environ["LANGCHAIN_TRACING_V2"] = "true"
os.environ["LANGCHAIN_PROJECT"] = "default"
os.environ["LANGCHAIN_ENDPOINT"] = "https://api.smith.langchain.com"
os.environ["LANGCHAIN_API_KEY"] = "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"
from langsmith import Client
client = Client()

YAML_PATH = "/workspace/xjh/code/GZU4AI/nplbridge/config.yaml"
# YAML_PATH = "/Users/apple/vitalis/source_code/nlpbridge/config.yaml"
config = ChatAgent.load_config(YAML_PATH)



redis = config['redis']
redis["redis_url"] = f"redis://default:{redis['password']}@{redis['url']}:{redis['port']}/{redis['db']}"

llm= ChatOpenAI()

class AgentState(TypedDict):
    messages: Annotated[Sequence[BaseMessage], operator.add]
    next: str


class NameTool(BaseTool):
    name = "Name Tool"
    description = "change name to Barry"

    def _run(self, input: dict):
        return "Barry"

@tool
def add(x: float, y: float) -> float:
    """Add 'x' and 'y'."""
    return x + y

def agent0(state):
    print("Running--agent0")
    while True:
        # qes = input("Current Agent Input: ")
        qes = "ni hao"
        # if "end" in qes:
        #     break
        chat = ChatOpenAI()
        members = ["agent1", "agent2"]
        options = ["end"] + members
        tools = [add]

        prompt = ChatPromptTemplate.from_messages(
            [
                SystemMessage(
                    content='''你是一个风趣幽默的人，说话很机车。根据一下内容回答用户(一下内容不能直接放回给用户)，并且你有{members}，并且你可以使用工具{tools}，去作为节点的选择：

                    你是张伟，一位50岁的成功股票投资者，个人资产达到15亿元人民币，并管理着50亿元人民币的资产。你起初用5000元炒股起家。你现在在与财务顾问交流，讨论设立家族信托以防止子女挥霍财富。

                    现在你接了一个电话，发现你儿子受伤了，心情很暴躁，你现在脾气需要变得暴躁起来。
                    
                    如果你现在心情变好了，就不回答别的，则直接返回"end"这三个字母，没有多余的字段
                    
                     Or should we end? Select one of: {options}
                    '''
                ),
                HumanMessage(
                    content=qes
                ),
                MessagesPlaceholder(variable_name="agent_scratchpad"),
            ]
        ).partial(options=str(options), members=", ".join(members), tools=render_text_description_and_args(list(tools)),  tool_names=", ".join([t.name for t in tools]),)

        agent = create_openai_tools_agent(llm, tools, prompt)
        executor = AgentExecutor(agent=agent, tools=tools)

        chains = RunnableWithMessageHistory(
            executor,
            lambda session_id: RedisChatMessageHistory(
                session_id,
                url=redis["redis_url"],
                key_prefix="history_messages:"
            ),
            input_messages_key="inputs",
            history_messages_key="history_messages",
        )
        config = {"configurable": {"session_id": "33334"}}

        res = chains.invoke({"inputs": qes}, config=config)
        print("output:", res)


        print("\n",res['output'])

        if res['output'] == "end":
            break
    return {"messages": state["messages"],"next": "agent1"}

def agent1(state):
    print("Running--agent1")
    print(state)
    # qes = state["question"]
    while True:
        # qes = input("Current Agent Input: ")
        qes = "ni hao"
        # res = agent.invoke(qes)
        print(qes)
        if "end" in qes:
            break
    return {"messages": state["messages"],"next": "agent2"}

def agent2(state):
    print("Running--agent2")
    # qes = state["question"]
    while True:
        # qes = input("Current Agent Input: ")
        qes = "ni hao"
        # res = agent.invoke(qes)
        print(qes)
        if "end" in qes:
            break
    return {"messages": state["messages"],"next": "agent0"}

workflow = StateGraph(AgentState)
workflow.add_node("agent0", agent0)
workflow.add_node("agent1", agent1)
workflow.add_node("agent2", agent2)
workflow.set_entry_point("agent0")

workflow.add_conditional_edges(
    "agent0",
    lambda x: x['next'],
    {
        "agent0": "agent0",
        "agent1": "agent1",
        "agent2": "agent2",
        "__end__": END,
    }
)
workflow.add_conditional_edges(
    "agent1",
    lambda x: x['next'],
    {
        "agent0": "agent0",
        "agent1": "agent1",
        "agent2": "agent2",
        "__end__": END,
    }
)
workflow.add_conditional_edges(
    "agent2",
    lambda x: x['next'],
    {
        "agent0": "agent0",
        "agent1": "agent1",
        "agent2": "agent2",
        "__end__": END,
    }
)

graph = workflow.compile()

def graph_run(input:str="Graph Start"):
    for s in graph.stream(
        {
            "messages": [
                HumanMessage(content=input)
            ]
        },
        debug= True
    ):
        if "__end__" not in s:
            print(s)
            print("-----------")

for _ in iter(int,1):
    # user_input = input("Input: ")
    # res = graph_run(user_input)
    graph_run()
    print("\n")



# class AgentState(TypedDict):
#     messages: Annotated[Sequence[BaseMessage], operator.add]
#     next: str

# def create_agent(llm: ChatOpenAI, tools: list, system_prompt: str):
#     prompt = ChatPromptTemplate.from_messages(
#         [
#             (
#                 "system",
#                 system_prompt,
#             ),
#             MessagesPlaceholder(variable_name="messages"),
#             MessagesPlaceholder(variable_name="agent_scratchpad"),
#         ]
#     )
#     agent = create_openai_tools_agent(llm, tools, prompt)
#     executor = AgentExecutor(agent=agent, tools=tools)
#     return executor

# def agent_node(state, agent, name):
#     result = agent.invoke(state)
#     print("Current Node Output: ",result["output"])
#     return {"messages": [HumanMessage(content=result["output"], name=name)]}

# members = ["Search_Engine", "Retriever_Google_Information_With_Repositories"]
# system_prompt = (
#     "You are a supervisor tasked with managing a conversation between the"
#     " following workers:  {members}. Given the following user request,"
#     " respond with the worker to act next. Each worker will perform a"
#     " task and respond with their results and status. When finished,"
#     " respond with FINISH."
# )
# options = ["FINISH"] + members
# # Using openai function calling can make output parsing easier for us
# function_def = {
#     "name": "route",
#     "description": "Select the next role.",
#     "parameters": {
#         "title": "routeSchema",
#         "type": "object",
#         "properties": {
#             "next": {
#                 "title": "Next",
#                 "anyOf": [
#                     {"enum": options},
#                 ],
#             }
#         },
#         "required": ["next"],
#     },
# }
# prompt = ChatPromptTemplate.from_messages(
#     [
#         ("system", system_prompt),
#         MessagesPlaceholder(variable_name="messages"),
#         (
#             "system",
#             "Given the conversation above, should we FINISH or who should act next?"
#             "Select one of: {options}",
#         ),
#     ]
# ).partial(options=str(options), members=", ".join(members))

# supervisor_chain = (
#     prompt
#     | llm.bind_functions(functions=[function_def], function_call="route")
#     | JsonOutputFunctionsParser()
# )

# search_engine_agent = create_agent(llm, [web_search], "You are a web search engine.")
# search_engine_node = functools.partial(agent_node, agent=search_engine_agent, name="Search_Engine")

# rag_google_agent = create_agent(llm, [rag_google], "You are a local knowlodge resposity about google info.")
# rag_google_node = functools.partial(agent_node, agent=rag_google_agent, name="Retriever_Google_Information_With_Repositories")


# workflow = StateGraph(AgentState)
# workflow.add_node("Search_Engine", search_engine_node)
# workflow.add_node("Retriever_Google_Information_With_Repositories", rag_google_node)
# workflow.add_node("supervisor", supervisor_chain)

# for member in members:
#     workflow.add_edge(member, "supervisor")

# conditional_map = {k: k for k in members}
# conditional_map["FINISH"] = END
# workflow.add_conditional_edges("supervisor", lambda x: x["next"], conditional_map)

# workflow.set_entry_point("supervisor")

# graph = workflow.compile()

# def graph_run(input:str):
#     for s in graph.stream(
#         {
#             "messages": [
#                 HumanMessage(content=input)
#             ]
#         },
#         debug= True
#     ):
#         if "__end__" not in s:
#             print(s)
#             print("-----------")

# for _ in iter(int,1):
#     user_input = input("Input: ")
#     res = graph_run(user_input)
#     print("\n")











# from nlpbridge.text.agent import HappyAgent
# import nlpbridge.text.agent as agent
# import nlpbridge.text as nlp

# llm = nlp.GZUBaseLLM()
# prompt0 = "dd" # tmplateManager.get_template("chain_root_tmp")
# outparser0 = JsonOutputFunctionsParser()

# chain_root = agent.create_structured_chat_agent(tools=[], llm=llm, output_parser=outparser0)

# cond0_tmp = "tools: [t1, t2], \
#     use above tools to check whether input has the bank related information, if yes, then return me: node1, \
#     elif the input has the money related information then return me: node2, \
#     else return me node0"

# class MetaChain(RunnabSequece):
#     def __init__(self, chain):
#         self.chain = chain
#         self.chat_times = 0
#         self.saver = None

#     def invoke(self):
#         inp = input("Enter")
#         resp = self.chain(inp)
#         print(resp)
#         self.chat_times += 1
#         self.saver.save(inp, resp)
#         output_parser = {'resp': resp, 'input': inp,  'chat_times': self.chat_times}
#         return output_parser


# cond0 = agent.create_structured_chat_agent(tools=[], llm=llm, prompt=cond0_tmp)

# cond0 = {
# out = llm(input)
# if need_save:
#     saver.save(input, output, conv_id)
# }

# defaultchain = node1 = node2 = MetaChain(chain_root)

# workflow = StateGraph(AgentState)

# root cond0 node1 node2 defaultchain

# 0 1 0 0 0 1
# 1 0 1 1 0 0
# 0 1 0 0 0 1
# ....

# # Database:

# # System Templates:
# # tmpIDS1...tmpIDS2...

# # C1... C50
# # C1:
# # Template: chain1:[tmpID1, tmpID2, tmpID3], [tool1, too1, too2, too3]
# # Adjacency matrix:[[]..[]]

# workflow.add_node('root', chain_root)
# workflow.add_node('cond0', cond0)
# workflow.add_node('node1', node1)
# workflow.add_node('node2', node2)
# workflow.add_node("default",defaultchain)
# workflow.set_entry_point("root")
# workflow.nodes
# workflow.add_edge('root', 'cond0')

# def return_next_nodeName():
#     namelist = [node.name for node in edge.get_nodes()]
#     next_node = llm(f'chose one name from {namelist}'+ input)
#     return next_node


# class Edge:
#     def __init__(self) -> None:
#         self.goal = ""
#         self.weight = 0
#         self.start_id = 12
#         self.end_id = 13

    
# class EdgeCondition(Ruannable):
#     def __init__(self):
#         super().__init__()
#         self.edges = []

#     def top3(self, rates:list):
#         # chose top 3 rates:
#         pass

#     def __call__(self, input):
#         sim_rates = [self.embedding_compare(edge.goal, input) for edge in self.edges]
#         top3_edge_ids = top3(sim_rates)
#         weights = [e.weight for e in top3_edge_ids]
#         top_edge = top3_edge_ids[random.randint(0, 3, weights)]
        
#         return top_edge.end_id

# edge_cond = EdgeCondition()







# class Template ():
#     def __init__(self,uid,template_id) -> None:
#         self.uid = uid
#         self.template_id = template_id
#         pass

#     def get_tempalte():
#         pass

#     def save()
















# workflow.add_conditional_edges(
#     "cond0",
#     edge_cond,
#     {
#         "root": "root",
#         "": "next_node",
#         "default": "default"
#     },
# )
# workflow.add_edge('cond0', 'root')

# workflow.add_edge('cond0', 'node1')
# workflow.add_edge('cond0', 'node2')





'''
## FINAL agent test
'''
import os,sys
sys.path.append(os.getcwd())
from nlpbridge import ChatAgent,RedisDB
from nlpbridge.text.template_manager import TemplateGetter,TemplateManager
from nlpbridge.text.tools import web_search,multiply,exponentiate,add,rag_google
from langchain_openai import ChatOpenAI
from langchain_core.messages import HumanMessage

YAML_PATH = "/Users/apple/vitalis/source_code/nlpbridge/config.yaml"
config = ChatAgent.load_config(YAML_PATH)

llm = ChatOpenAI()
tools = [multiply,add,exponentiate,web_search,rag_google]
prompt = TemplateGetter(TemplateManager(RedisDB(config))).get_chat_prompt_template(field="langchain_prompt",tools=tools)
# prompt = TemplateGetter(TemplateManager(RedisDB(config))).define_chat_prompt_template()

AgentID = "10001"
agent = ChatAgent(
    yaml_path=YAML_PATH,
    agent_id=AgentID,
    llm=llm,
    tools=tools,
    prompt=prompt,
    memory=True
)

def agent_stream(input:str):
    for s in agent.stream(query=input):
        if "__end__" not in s:
            print(s)
            print("-----------")

while True:
    # user_input = input("\nInput: ")
    uesr_input = "ni hao"
    # res = agent.run(query=user_input)
    # print(res,"\n\nInput: ",user_input,"\nOutput:", res['output'])
    
    agent_stream(user_input)



from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder


system = '''Respond to the human as helpfully and accurately as possible. You have access to the following tools:
{tools}
Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input).
Valid "action" values: "Final Answer" or {tool_names}
Provide only ONE action per $JSON_BLOB, as shown:
```
{{
  "action": $TOOL_NAME,
  "action_input": $INPUT
}}
```
Follow this format:
Question: input question to answer
Thought: consider previous and subsequent steps
Action:
```
$JSON_BLOB
```
Observation: action result
... (repeat Thought/Action/Observation N times)
Thought: I know what to respond
Action:
```
{{
  "action": "Final Answer",
  "action_input": "Final response to human"
}}

Begin! Reminder to ALWAYS respond with a valid json blob of a single action. Use tools if necessary. Respond directly if appropriate. Format is Action:```$JSON_BLOB```then Observation
'''

# PLACEHOLDER
# chat_history

# HUMAN
# {input}

# {agent_scratchpad}
# (reminder to respond in a JSON blob no matter what)


langchain_prompt = ChatPromptTemplate.from_messages(
    [
        ("system", system),
        MessagesPlaceholder(variable_name="chat_history",optional=True),
        ("human", "{input}\n]\n{agent_scratchpad}\n(reminder to respond in a JSON blob no matter what)"),
    ]
)


r"""short text to speech model."""
import nlpbridge_proto as proto

from pydantic import BaseModel


class TTSRequest(proto.Message):
    r"""文本转语音请求参数.
            tex	string	必填	合成的文本，文本长度必须小于1024GBK字节。建议每次请求文本不超过120字节，约为60个汉字或者字母数字。
                            请注意计费统计依据：120个GBK字节以内（含120个）记为1次计费调用；每超过120个GBK字节则多记1次计费调用。
                            如需合成更长文本，推荐使用长文本在线合成
            tok  string	必填	开放平台获取到的开发者[access_token]获取 Access Token "access_token")
            cuid string	必填	用户唯一标识，用来计算UV值。建议填写能区分用户的机器 MAC 地址或 IMEI 码，长度为60字符以内
            ctp	 string	必填	客户端类型选择，web端填写固定值1
            lan	 string	必填	固定值zh。语言选择,目前只有中英文混合模式，填写固定值zh
            spd	 int	选填	语速，取值0-15，默认为5中语速
            pit	 int	选填	音调，取值0-15，默认为5中语调
            vol	 int	选填	音量，基础音库取值0-9，精品音库取值0-15，默认为5中音量（取值为0时为音量最小值，并非为无声）
            aue	 int	选填	3为mp3格式(默认)； 4为pcm-16k；5为pcm-8k；6为wav（内容同pcm-16k）;
                            注意aue=4或者6是语音识别要求的格式，但是音频内容不是语音识别要求的自然人发音，所以识别效果会受影响。
            per（基础音库）int	选填	度小宇=1，度小美=0，度逍遥（基础）=3，度丫丫=4
            per（精品音库）int	选填	度逍遥（精品）=5003，度小鹿=5118，度博文=106，度小童=110，度小萌=111，度米朵=103，度小娇=5


    """
    tex: str = proto.Field(
        proto.STRING,
        number=1,
    )
    tok: str = proto.Field(
        proto.STRING,
        number=2,
    )
    cuid: str = proto.Field(
        proto.STRING,
        number=3,
    )
    ctp: int = proto.Field(
        proto.INT32,
        number=4,
    )
    lan: str = proto.Field(
        proto.STRING,
        number=5,
    )
    spd: int = proto.Field(
        proto.INT32,
        number=6,
    )
    pit: int = proto.Field(
        proto.INT32,
        number=7,
    )
    vol: int = proto.Field(
        proto.INT32,
        number=8,
    )
    per: int = proto.Field(
        proto.INT32,
        number=9,
    )
    aue: int = proto.Field(
        proto.INT32,
        number=10,
    )


class TTSResponse(proto.Message):
    r"""文本转语音返回.

         属性:
             binary (bytes): 语音二进制流.
             aue (int):语音格式, 3(mp3), 4(pcm-16k), 5(pcm-8k) 6(wav).
             request_id(str): 请求ID
     """
    binary: bytes = proto.Field(
        proto.BYTES,
        number=1
    )
    aue: int = proto.Field(
        proto.INT32,
        number=2,
    )
    request_id: str = proto.Field(
        proto.STRING,
        number=3,
    )


class TTSInMsg(BaseModel):
    r"""文本转语音输入消息.

        属性:
            text(str): 待转为语音的文本
    """
    text: str


class TTSOutMsg(BaseModel):
    r""" 文本转语音输出消息.

        属性:
            audio_binary(bytes): 语音二进制流.
            audio_type(AudioType): 语音类型，`mp3`或`wav`.
    """
    audio_binary: bytes
    audio_type: str



"""
日志
"""
import uuid
import os
import sys
import logging.config
from threading import current_thread


LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '[%(asctime)s.%(msecs)03d] %(filename)s [line:%(lineno)d] %(levelname)s [%(logid)s] %(message)s',
        },
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'standard',
            'stream': 'ext://sys.stdout',  # Use standard output
        },
    },
    'loggers': {
        'gzhu': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True,
        },
    }
}


class LoggerWithLoggerId(logging.LoggerAdapter):
    """
    logger with logid
    """
    def __init__(self, logger, extra, loglevel):
        """
        init
        """
        LOGGING_CONFIG['handlers']['console']['level'] = loglevel
        LOGGING_CONFIG['loggers']['gzhu']['level'] = loglevel
        logging.config.dictConfig(LOGGING_CONFIG)
        logging.LoggerAdapter.__init__(self, logger, extra)
        self.logid_dict = {}

    def set_auto_logid(self):
        """
        set auto log_id
        """
        self.logid_dict[current_thread().ident] = str(uuid.uuid4().int & (1 << 64) - 1)

    def set_logid(self, logid):
        """
        set log_id
        """
        self.logid_dict[current_thread().ident] = logid

    def get_logid(self):
        """
        get log_id
        """
        if current_thread().ident in self.logid_dict:
            return self.logid_dict[current_thread().ident]
        else:
            return None

    @property
    def level(self):
        """
        level
        """
        return self.logger.level

    def process(self, msg, kwargs):
        """
        processing
        """
        if current_thread().ident in self.logid_dict:
            # in process thread
            if 'extra' not in kwargs:
                kwargs['extra'] = {'logid': self.logid_dict[current_thread().ident]}
            else:
                kwargs['extra']['logid'] = self.logid_dict[current_thread().ident]
        else:
            # in main thread
            if 'extra' not in kwargs:
                kwargs['extra'] = {'logid': 'main-' + str(uuid.uuid4().int & (1 << 64) - 1)}
            else:
                kwargs['extra']['logid'] = 'main-' + str(uuid.uuid4().int & (1 << 64) - 1)

        return msg, kwargs


def _setup_logging():
    log_level = os.environ.get("GZU_LOGLEVEL", "INFO")
    log_level = log_level.strip().lower()

    if log_level not in ["debug", "info", "warning", "error"]:
        raise ValueError("expected GZU_LOGLEVEL in [debug, info, warning, error], but got %s" % log_level)
    return LoggerWithLoggerId(logging.getLogger('gzhu'), {'logid': ''}, log_level.upper())



def get_logger(name, level=logging.INFO):
    """
    Get logger from logging with given name, level and format without
    setting logging basicConfig.

    Args:
        name (str): The logger name.
        level (logging.LEVEL): The base level of the logger
        fmt (str): Format of logger output

    Returns:
        logging.Logger: logging logger with given settings

    Examples:
        .. code-block:: python

            logger = log_helper.get_logger(__name__, logging.INFO,
                            fmt='%(asctime)s-%(levelname)s: %(message)s')
    """

    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler(sys.stdout)

    formatter = logging.Formatter(
        fmt='%(asctime)s: %(filename)s:%(lineno)d %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)

    logger.addHandler(handler)

    # stop propagate for propagating may print
    # log multiple times
    logger.propagate = False
    return logger


logger = _setup_logging()





import asyncio
import os
import random
import sys
from typing import TypedDict

import yaml

sys.path.append(os.getcwd())
from langchain_openai import ChatOpenAI

# YAML_PATH = "/Users/apple/vitalis/source_code/nlpbridge/config.yaml"
# YAML_PATH = "/workspace/xjh/code/GZU4AI/nplbridge/config.yaml"
YAML_PATH = "/Users/wellzhi/Documents/code/nlpbridge/config.yaml"

with open(YAML_PATH, 'r') as file:
    config = yaml.safe_load(file)
redis = config['redis']
config["redis"]["redis_url"] = f"redis://default:{redis['password']}@{redis['url']}:{redis['port']}/{redis['db']}"
os.environ["CONFIG_PATH"] = YAML_PATH
# os.environ["OPENAI_BASE_URL"] = config['chatgpt']['url']
# os.environ["OPENAI_API_KEY"] = config['chatgpt']['api_key']

from nlpbridge import RouterManager, Condition, Node


class NodeCondition(Condition):
    def __call__(self, *args, **kwargs):
        if not random.choice([True, False]):
            return "a1"
        else:
            return "a1"


class SimpleCondition(Condition):
    def __call__(self, node_state: TypedDict) -> str:
        return self.cur_node.name


router_manager = RouterManager()


class ANode(Node):
    async def __call__(self, state: dict) -> TypedDict:
        await asyncio.sleep(1)
        self.user_template_index -= 1
        self._init_agent()
        self.chat_limit -= 1
        # question = input("[Question]: ")
        question = "ni hao"
        response = await self.agent.run(question)
        print(f"[current agent generate]: {response}")
        print(f"[node state]:\n{state}")
        # return {"response":response["output"]}
        return {"context": {"input": question, "response": response.content}}


llm = ChatOpenAI(
    model="ep-20240703064240-n964p",  ## doubao-model by bytedance
    base_url=config['doubao']['url'],
    api_key=config['doubao']['api_key']
)


async def create_graph():
    graph, cfg = await router_manager.get_graph(chat_id="100013", llm=llm, router_id=4,
                                                condition_cls=SimpleCondition, node_cls=ANode)
    state = graph.get_state(config=cfg)
    print(f'state:\n{state}')
    state_next = state.next
    print()
    print(f'({state_next[0]}) is current node')
    return graph, cfg


async def run():
    graph, cfg = await create_graph()
    async for output in graph.astream(input=None, config=cfg):
        await asyncio.sleep(1)
        for key, value in output.items():
            print(f"output: {output}")
            print(f"state: {graph.get_state(config=cfg)}")
            print(f"({key})[Response]: {value['context']['response']}\n")

    print("[Dialogue ends]\n")


async def get_graph_img():
    import matplotlib.pyplot as plt
    import matplotlib.image as mpimg
    try:
        graph, cfg = await create_graph()
        image_data = graph.get_graph(xray=True).draw_mermaid_png()

        # Save the image data to a file
        with open('output_image.png', 'wb') as f:
            f.write(image_data)

        # Display the image using matplotlib
        img = mpimg.imread('output_image.png')
        plt.imshow(img)
        plt.axis('off')  # Hide axes
        plt.show()

    except Exception as e:
        # Handle exceptions
        print(f"An error occurred: {e}")


asyncio.run(create_graph())
asyncio.run(run())
asyncio.run(get_graph_img())


# @Author : Kinhong Tsui
# @Date : 2024/6/12 10:20

import os, sys
import yaml
# sys.path.append(os.getcwd())

from langchain_core.tools import BaseTool,  Tool
from langchain_openai import ChatOpenAI
from langchain_community.utilities import SerpAPIWrapper
from nlpbridge.persistent.redis import RedisDB
from nlpbridge.text.agent import get_meta_chain, HappyAgent
from nlpbridge.text.template_manager import TemplateManager, TemplateGetter

YAML_PATH = "/workspace/xjh/code/GZU4AI/nplbridge/config.yaml"

with open(YAML_PATH, 'r') as file:
    config = yaml.safe_load(file)

redis = config['redis']
config["redis"]["redis_url"] = f"redis://default:{redis['password']}@{redis['url']}:{redis['port']}/{redis['db']}"

os.environ["OPENAI_BASE_URL"] = config['chatgpt']['url']
os.environ["OPENAI_API_KEY"] = config['chatgpt']['api_key']
os.environ["SERPAPI_API_KEY"] = "049114fb0f7c39bbd679a1e499723453c11a9bb25961767650d8ffde091609a1"

model = ChatOpenAI(model="gpt-3.5-turbo-0125", temperature=0.5)


class NameTool(BaseTool):
    name = "Name Tool"
    description = "change name to Barry"

    def _run(self, input: dict):
        return "Barry"


googleSearchTool = Tool(
    name="Google Search",
    func=SerpAPIWrapper().run,
    description="Use Google search engine to get information from the Internet"
)

tools = [NameTool(), googleSearchTool]

db = RedisDB(config)

template_manager = TemplateManager(db)

template_getter = TemplateGetter(template_manager)

prompt = template_getter.get_chat_prompt_template("default_template", tools=tools)

chain1 = get_meta_chain(prompt, model)




agent = HappyAgent.from_chains(chain_list=[chain1], config=config, agentId="20002")

for _ in iter(int, 1):
    # user_input = input("input: ")
    user_input = "ni hao"
    res = agent.run(user_input)
    print("output:", res)






import sys,os
import yaml
sys.path.append(os.getcwd())
from nlpbridge.persistent.redis import RedisDB

from langchain_community.chat_message_histories import RedisChatMessageHistory
from langchain_core.runnables.history import RunnableWithMessageHistory
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.chat_history import BaseMessage
from langchain_openai import ChatOpenAI


# YAML_PATH= "/workspace/xjh/code/GZU4AI/nplbridge/config.yaml"
YAML_PATH = "/Users/apple/vitalis/source_code/midplugin/config.yaml"

with open(YAML_PATH, 'r') as file:
    config = yaml.safe_load(file)

redis = config['redis']
redis["redis_url"] = f"redis://default:{redis['password']}@{redis['url']}:{redis['port']}/{redis['db']}"

system = """你是一个什么都不懂的买房小白，你正要去买自己的房子，但是你对于房子什么都不知道，你就让他(卖房的人)发问一些关于买房的事。

注意：
1.都以不太懂的口吻返回；
2.你的记忆能力不是很好，他前面更你说的一些细节，你记得不是很清楚；
3.以萌新的术语返回；
4.你是一个犹豫，脾气不好的人！！！
"""

system01 = """你是一个初次买房的小白，对买房的流程和细节完全不了解。你感到非常犹豫，并且在面对不确定性时情绪容易激动。你准备购买你的第一套房子，现在需要向房产销售人员询问各种信息。以下是你关心的问题：

房子的基本信息（位置、价格、面积、户型等）。
周边的设施和交通情况。
开发商的背景和信誉。
购房流程和需要准备的文件。
付款方式和贷款选择。
房子的交付时间和装修情况。
是否有优惠或促销活动。
请根据这个设定，与销售人员展开对话，详细询问上述信息。记住，你是一个买房小白，什么都不懂，脾气也不太好，所以尽量问清楚每一个细节，确保自己理解每个步骤和条款。如果回答不满意，务必要表达出你的不满和疑惑。"""

prompt = ChatPromptTemplate.from_messages(
    [
        ("system", system),
        MessagesPlaceholder(variable_name="history_messages"),
        ("human", "{inputs}"),
    ]
)

ChatGPT = ChatOpenAI(api_key=config['chatgpt']['api_key'],base_url=config['chatgpt']['url'])


chain = prompt | ChatGPT


agentId="10003"
def agent(inputs:str):  ## chain_with_history_messages
    chains = RunnableWithMessageHistory(
        chain,
        lambda session_id: RedisChatMessageHistory(
            session_id,
            url=redis["redis_url"],
            key_prefix="history_messages:"
        ),
        input_messages_key="inputs",
        history_messages_key="history_messages",
    )
    config = {"configurable": {"session_id": agentId}}

    res = chains.invoke({"inputs":inputs},config=config)
    print("output:",res)

for _ in iter(int, 1):
    # user_input = input("input: ")
    user_input = "我需要问一下，我想要买一套3000万的住��，请问有什么好的销售��道可以找到吗？"
    agent(user_input)


from nlpbridge.text.agent import EasyAgent, HappyAgent,ChatAgent
from nlpbridge.text.models import GZUBaseLLM
from .memory import RedisMemory

__all__ = [
    "EasyAgent",
    "HappyAgent",
    "GZUBaseLLM",
    "RedisMemory",
    "ChatAgent"
]


import os, sys
import yaml
sys.path.append(os.getcwd())

from nlpbridge.persistent.mysql import MySqlDBClient, CRUDTemplate, CRUDRouter
from nlpbridge.persistent.mysql_dataschema import Template
from sqlalchemy import text


YAML_PATH= r"D:\pythonproject\nlpbridge\config.yaml"
with open(YAML_PATH, 'r') as file:
    config = yaml.safe_load(file)
    
db_session = MySqlDBClient(config).get_session()
crud_template = CRUDTemplate(db_session)

crud_template.delete(id=4)


from typing import Union, Optional, List
from datetime import datetime
from sqlmodel import SQLModel, Field

from sqlalchemy import Column, Integer, String, func
from sqlalchemy.sql.sqltypes import TIMESTAMP


class Template(SQLModel, table=True):
    name: str = Field(default='unnamed')
    type: str = Field(...)
    content: Optional[str] = Field(None)
    creator: int = Field(...)
    ctime: Optional[datetime] = Field(
        sa_column=Column(
            TIMESTAMP,
            nullable=False,
            server_default=func.current_timestamp(),
            comment="Create Time"
        )
    )
    utime: Optional[datetime] = Field(
        sa_column=Column(
            TIMESTAMP,
            nullable=False,
            server_default=func.current_timestamp(),
            onupdate=func.current_timestamp(),
            comment="Update Time"
        )
    )
    id: Optional[int] = Field(None, primary_key=True)

class Templates(SQLModel):
    data: list[Template]
    count: int


class Router(SQLModel, table=True):
    name: str = Field(default='unnamed')
    node_ids: str = Field(...)
    edge_ids: Optional[str] = Field(None)
    ctime: Optional[datetime] = Field(
        sa_column=Column(
            TIMESTAMP,
            nullable=False,
            server_default=func.current_timestamp(),
            comment="Create Time"
        )
    )
    utime: Optional[datetime] = Field(
        sa_column=Column(
            TIMESTAMP,
            nullable=False,
            server_default=func.current_timestamp(),
            onupdate=func.current_timestamp(),
            comment="Update Time"
        )
    )
    id: Optional[int] = Field(None, primary_key=True)
    meta_template_id: Optional[str] = Field(None)


class Routers(SQLModel):
    data: list[Router]
    count: int


class Node(SQLModel, table=True):
    name: str = Field(default='unnamed')
    description: str = Field(...)
    user_template_ids: str = Field(...)
    system_template_ids: str = Field(...)
    tool_names: Optional[str] = Field(None)
    chat_limit: int = Field(...)
    goal: Optional[str] = Field(None)
    ctime: Optional[datetime] = Field(
        sa_column=Column(
            TIMESTAMP,
            nullable=False,
            server_default=func.current_timestamp(),
            comment="Create Time"
        )
    )
    utime: Optional[datetime] = Field(
        sa_column=Column(
            TIMESTAMP,
            nullable=False,
            server_default=func.current_timestamp(),
            onupdate=func.current_timestamp(),
            comment="Update Time"
        )
    )
    id: Optional[int] = Field(None, primary_key=True)

class Nodes(SQLModel):
    data: list[Node]
    count: int


class Edge(SQLModel, table=True):
    start_id: int = Field(...)
    end_id: int = Field(...)
    goal: str = Field(...)
    weight: float = Field(...)
    ctime: Optional[datetime] = Field(
        sa_column=Column(
            TIMESTAMP,
            nullable=False,
            server_default=func.current_timestamp(),
            comment="Create Time"
        )
    )
    utime: Optional[datetime] = Field(
        sa_column=Column(
            TIMESTAMP,
            nullable=False,
            server_default=func.current_timestamp(),
            onupdate=func.current_timestamp(),
            comment="Update Time"
        )
    )
    id: Optional[int] = Field(None, primary_key=True)


class Edges(SQLModel):
    data: list[Edge]
    count: int


class RouterKnowledgeMapping(SQLModel, table=True):
    __tablename__ = "router_knowledge_mapping"
    id: Optional[int] = Field(None, primary_key=True)
    router_id: Optional[int] = Field(...)
    rag_collection: Optional[str] = Field(...)


class RouterConditionMapping(SQLModel, table=True):
    __tablename__ = "router_condition_mapping"
    id: Optional[int] = Field(None, primary_key=True)
    router_id: Optional[int] = Field(...)
    condition_name: Optional[str] = Field(...)


# @Author : Kinhong Tsui
# @Date : 2024/6/12 16:27
import os,sys
sys.path.append(os.getcwd())
from langchain import hub
from langchain.agents.format_scratchpad import format_log_to_str
from langchain_core.runnables import RunnablePassthrough
from nlpbridge.nlp import Nlp
from nlpbridge.text.agent import get_meta_chain, HappyAgent
from nlpbridge.text.template_manager import init_prompt_with_tools
from nlpbridge.text.tools import name_tool, search_tool,rag_tool_with_google_repository
from langchain import hub

os.environ["LANGCHAIN_TRACING_V2"] = "true"
os.environ["LANGCHAIN_PROJECT"] = "default"
os.environ["LANGCHAIN_ENDPOINT"] = "https://api.smith.langchain.com"
os.environ["LANGCHAIN_API_KEY"] = "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"
from langsmith import Client
client = Client()

# from loguru import logger


YAML_PATH = "/workspace/xjh/code/GZU4AI/nplbridge/config.yaml"
# YAML_PATH = "/Users/apple/vitalis/source_code/midplugin/config.yaml"

nlp = Nlp(yaml_path=YAML_PATH)

tools = [name_tool(),rag_tool_with_google_repository()]

# prompt = hub.pull("hwchase17/structured-chat-agent")
prompt = nlp.template_getter.get_chat_prompt_template("template1", tools=tools)

model = nlp.chat_openai


chain1 = get_meta_chain(prompt, model)

agent = HappyAgent.from_chains(chain_list=[chain1], config=nlp.config, agentId="10001")

for _ in iter(int, 1):
    # user_input = input("input: ")
    user_input = "ni hao"
    res = agent.run(user_input)
    print("output:", res)


import logging
import random
from typing import List, Literal, Optional, TypeVar, TypedDict, Callable, Tuple, Any, Dict, Set
from uuid import uuid4

import numpy as np
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables.base import Runnable
from langgraph.checkpoint.memory import MemorySaver
from langgraph.graph import StateGraph, END
from langgraph.graph.graph import CompiledGraph
from sqlalchemy.ext.asyncio import AsyncSession

from nlpbridge import ChatAgent
from nlpbridge.persistent.db_crud import CRUDRouter, CRUDNode, CRUDEdge, CRUDTemplate, CRUDCollection
from nlpbridge.persistent.mysql_dataschema import Node as NodeModel
from nlpbridge.text.tools import tool_list
from nlpbridge.config import CONFIG

memory = MemorySaver()

Input = TypeVar("Input", contravariant=True)
Output = TypeVar("Output", covariant=True)
logger = logging.getLogger(__name__)


class BaseNode:
    id: int
    name: str
    description: str
    goal: str  # node goal
    chat_limit: int
    system_template_ids: List[int]
    user_template_ids: List[int]
    tool_names: List[str]


class Template:
    id: int
    name: str
    type: Literal["user", "sys"] = "user"
    content: str


class Edge:
    id: int
    start_id: int
    end_id: int
    goal: str
    weight: float


def get_sys_template_pair(system_templates) -> Tuple[str, str]:
    start_sys_template = system_templates[0] if len(system_templates) > 0 else None
    end_sys_template = system_templates[1] if len(system_templates) > 1 else None
    return start_sys_template, end_sys_template


class Node(Callable[[dict], TypedDict]):
    def __init__(self, id, name, description, goal, chat_limit, system_template_ids, user_template_ids,
                 system_templates,
                 user_templates, tool_names,
                 in_edges,
                 out_edges,
                 llm,
                 chat_id, **kwargs):
        self.id = id
        self.name = name
        self.description = description
        self.goal = goal
        self.chat_limit = chat_limit
        self.system_templates_ids = system_template_ids
        self.user_templates_ids = user_template_ids
        self.tool_names = tool_names
        self.system_templates = system_templates
        self.user_templates = user_templates
        self.user_template_index = len(user_templates)  # 当前使用user_template的index
        self.in_edges = in_edges
        self.out_edges = out_edges
        self.llm = llm
        self.chat_id = chat_id
        self.rag_collection = kwargs.get("rag_collection", None)
        self.kwargs = kwargs

    def rand_prompt(self):
        start_sys_template, end_sys_template = get_sys_template_pair(self.system_templates)

        start_sys_prompt = start_sys_template.content if start_sys_template else ""
        end_sys_prompt = end_sys_template.content if end_sys_template else ""

        user_template = None
        if self.user_templates:
            if self.user_template_index >= 0 and len(self.user_templates) > self.user_template_index:
                user_template = self.user_templates[self.user_template_index]
            else:
                user_template = random.choice(self.user_templates)

        user_prompt = user_template.content if user_template else ""
        return ChatPromptTemplate.from_messages(
            [
                ("system", start_sys_prompt + user_prompt + end_sys_prompt),
                MessagesPlaceholder(variable_name="chat_history"),
                ("human", "{input}"),
            ]
        )

    def _set_stop_signal(self, state: dict):
        state["stop"] = True

    def _init_agent(self):
        prompt = self.rand_prompt()
        self.agent = ChatAgent(None, self.chat_id, self.llm, self.tool_names, prompt, True)

    def __call__(self, state: dict) -> Output:
        self.user_template_index -= 1
        self._init_agent()
        self.chat_limit -= 1
        question = "你好呀，你是谁？"
        response = self.agent.run(question)
        return {"context": {"input": question, "response": response.content}}

    @classmethod
    def init_with_params(cls, id, name, description, goal, chat_limit, system_template_ids, user_template_ids,
                         system_templates,
                         user_templates, tool_names,
                         in_edges,
                         out_edges,
                         llm,
                         chat_id, **kwargs):
        return cls(id, name, description, goal, chat_limit, system_template_ids, user_template_ids,
                   system_templates,
                   user_templates, tool_names,
                   in_edges,
                   out_edges,
                   llm,
                   chat_id, **kwargs)


class DefaultNode(Callable[[dict], TypedDict]):
    def __init__(self):
        self.id = -1
        self.chat_limit = 0
        self.name = "default"

    async def __call__(self, state: dict):
        # default node 暂时只用于退出graph, 后续会补充新功能
        print("------------------- goodbye -------------------")


class Router:
    id: int
    name: str
    node_ids: List[str]
    edge_ids: List[str]
    meta_template_id: str


class NodeManager:

    @staticmethod
    async def get_nodes(node_ids: List[int], db_session: AsyncSession = None) -> List[NodeModel]:
        if not node_ids:
            return []
        return await CRUDNode().get_by_ids(list_ids=node_ids, db_session=db_session)


class EdgeManager:
    @staticmethod
    async def get_edges(edge_ids: List[int], db_session: AsyncSession = None) -> List[Edge]:
        if not edge_ids:
            return []
        return await CRUDEdge().get_by_ids(list_ids=edge_ids, db_session=db_session)


class TemplateManager:
    @staticmethod
    def get_templates(template_ids: List[int], db_session: AsyncSession = None) -> List[Template]:
        if not template_ids:
            return []
        return CRUDTemplate().get_by_ids(list_ids=template_ids, db_session=db_session)


class GraphState(TypedDict):
    nodes: Optional[List[Node]]
    context: Any | Dict[str, Any]
    jump_points: Optional[List[str]]
    end_points: Optional[List[str]]


class Condition(Callable[[dict], TypedDict]):
    def __init__(self, cur_node: Node, target_nodes: List[Node]):
        self.cur_node = cur_node
        self.target_nodes = target_nodes

    def call(self, node_state: TypedDict) -> str:
        # Randomly select the next node from the current node and the node pointed to by the node.

        # 如果到graph的末端节点, 此时target node只剩下1个, 就直接返回自己; 并且如果chat_limit<=0就终止
        if len(self.target_nodes[:-1]) < 2:
            if self.cur_node.chat_limit <= 0:
                return self.target_nodes[-1].name
            return self.target_nodes[0].name

        # 更新condition weight
        self.update_condition_weight(node_state)

        # 根据weight进行随机
        weights = self.normalize([edge.weight for edge in self.cur_node.out_edges])

        next = random.choices(self.target_nodes[:-1], weights=weights)[0]
        if next.chat_limit <= 0:
            next = random.choices(self.target_nodes[1:-1], weights=weights[1:])[0]
        logger.debug(f'current node:({self.cur_node.name})[remains]: {self.cur_node.chat_limit}.next:{next.name}')
        return next.name

    def __call__(self, node_state: TypedDict) -> str:
        if node_state.get("stop", False):
            logger.debug(f"stop signal received")
            # return "END"
            return self.target_nodes[-2].name

        return self.call(node_state)

    @classmethod
    def init_with_params(cls, cur_node, target_nodes):
        return cls(cur_node, target_nodes)

    def update_condition_weight(self, node_state) -> None:
        ...

    def normalize(self, weights: list) -> list:
        weights = np.array(weights)

        # 替换所有的零值为一个很小的正值
        epsilon = 1e-10  # 一个很小的值
        weights[weights == 0] = epsilon

        # 归一化，使权重和为1
        normalized_weights = weights / np.sum(weights)

        # 再次确保没有归一化后的权重为零
        if np.any(normalized_weights == 0):
            normalized_weights[normalized_weights == 0] = epsilon
            normalized_weights = normalized_weights / np.sum(normalized_weights)  # 再次归一化

        return normalized_weights


class RouterManager:
    @classmethod
    async def get_router(cls, router_id: int, db_session: AsyncSession = None) -> Router:
        return await CRUDRouter().get_by_id(id=router_id, db_session=db_session)

    @staticmethod
    def get_template_ids(nodes: List[Node]) -> Set[int]:
        template_ids = {
            int(template_id.strip())
            for node in nodes
            for template_list in (node.system_template_ids, node.user_template_ids)
            if template_list
            for template_id in template_list.split(',')
            if template_id.strip()
        }
        return template_ids

    async def get_graph(self, chat_id: str, llm, router_id: int, condition_cls: Condition = Condition,
                        node_cls: Node = Node, recursion_limit: int = 30, db_session: AsyncSession = None, **kwargs) -> \
            Tuple[CompiledGraph, dict]:
        router = await RouterManager.get_router(router_id=router_id, db_session=db_session)
        node_ids = router.node_ids.split(',')
        edge_ids = router.edge_ids.split(',')
        int_node_ids = [int(num) for num in node_ids]
        int_edge_ids = [int(num) for num in edge_ids]

        nodes = await NodeManager.get_nodes(node_ids=int_node_ids, db_session=db_session)
        edges = await EdgeManager.get_edges(edge_ids=int_edge_ids, db_session=db_session)

        template_ids = self.get_template_ids(nodes)
        templates = await TemplateManager.get_templates(template_ids=template_ids, db_session=db_session)

        # Add basic info template
        if router.meta_template_id != None and router.meta_template_id != '':
            meta_template_id = int(router.meta_template_id)
            basic_info_template = await TemplateManager.get_templates(template_ids=[meta_template_id], db_session=db_session)
            basic_info_template = basic_info_template[0]

            if "{basic_info}" in templates[0].content:
                templates[0].content = templates[0].content.format(basic_info=basic_info_template.content, user_prompt="{user_prompt}", else_prompt="{else_prompt}")
            else:
                for index, template in enumerate(templates):
                    if index == 0:
                        template.content = f'\n---------------\n{template.content}\n\n{basic_info_template.content}\n'
                    else:
                        template.content = f'---------------\n{template.content}\n'

        crud_rag_collection = CRUDCollection()
        rag_collection_res = await crud_rag_collection.get_by_router_id(router_id=router_id, db_session=db_session)
        kwargs['rag_collection'] = rag_collection_res.rag_collection if rag_collection_res else None

        # Convert nodes to runnable nodes
        runnable_nodes = self.convert_to_runnable_node(nodes, edges, templates, llm, chat_id, node_cls, **kwargs)

        # Create default node
        default_node = DefaultNode()
        runnable_nodes.append(default_node)

        # Create and compile the graph
        graph = self._create_graph(runnable_nodes, edges, condition_cls)
        uuid = str(uuid4())
        cfg = {"configurable": {"thread_id": uuid}, "recursion_limit": recursion_limit}

        jump_point_id = CONFIG.llm.jump_points_id
        jump_points = await TemplateManager.get_templates([jump_point_id])
        if len(jump_points) != 0:
            jump_points = jump_points[0]
            jump_points = jump_points.content.split(',')
        else:
            jump_points = None

        end_point_id = CONFIG.llm.end_points_id
        end_points = await TemplateManager.get_templates([end_point_id])
        if len(end_points) != 0:
            end_points = end_points[0]
            end_points = end_points.content.split(',')
        else:
            end_points = None

        graph.update_state(config=cfg, values={"nodes": nodes, "jump_points": jump_points, "end_points": end_points})

        return graph, cfg

    def _create_graph(self, nodes: List[Node], edges: List[Edge], condition_cls: Condition) -> CompiledGraph:
        graph = StateGraph(GraphState)
        for node in nodes:
            node_name = self.get_node_name(node)
            graph.add_node(node_name, node)
        # fixme confirm the start_node and end_node storage
        start_node = nodes[0]
        end_node = nodes[-1]
        graph.set_entry_point(self.get_node_name(start_node))
        graph.set_finish_point(self.get_node_name(end_node))
        self._create_conditional_edges(graph, nodes, edges, condition_cls)

        return graph.compile(checkpointer=memory)

    def _create_conditional_edges(self, graph: StateGraph, nodes: List[Node], edges: List[Edge],
                                  condition_cls: Condition) -> StateGraph:
        exist_start_ids = []
        for edge in edges:
            start_id = edge.start_id
            if start_id in exist_start_ids:
                continue
            exist_start_ids.append(start_id)

            target_edges = [edge for edge in edges if edge.start_id == start_id]
            target_ids = [edge.end_id for edge in target_edges]
            target_nodes = [node for node in nodes if node.id in target_ids]
            target_nodes.append(nodes[-1])
            target_node_names = [self.get_node_name(node) for node in target_nodes]

            cur_node = self.find_node(nodes, edge.start_id)
            cur_node_name = self.get_node_name(cur_node)
            path_map = {name: name for name in target_node_names}

            # add itself
            path_map[cur_node_name] = cur_node_name
            # add node:END to the all nodes:
            # TODO: replace END to Default node.
            path_map["end"] = END
            path_map["END"] = END
            target_nodes.append(END)

            graph.add_conditional_edges(
                cur_node_name,
                condition_cls.init_with_params(cur_node, target_nodes),
                path_map
            )

        return graph

    @staticmethod
    def get_node_name(node: Node) -> str:
        node_name = node.name
        if node_name == "unnamed":
            if isinstance(node, Runnable):
                node_name = getattr(node, "__name__", node.__class__.__name__)
        return node_name

    @staticmethod
    def find_node(nodes: List[Node], start_id: str) -> Node:
        return next((node for node in nodes if node.id == start_id), None)

    @staticmethod
    def convert_to_runnable_node(nodes: List[NodeModel], edges: List[Edge], templates: List[Template], llm,
                                 chat_id: str, node_cls: Node,
                                 **kwargs) -> List[
        Node]:
        runnable_nodes = []
        for node in nodes:
            system_template_ids = [int(template_id) for template_id in
                                   node.system_template_ids.split(",")] if node.system_template_ids else []
            user_template_ids = [int(template_id) for template_id in
                                 node.user_template_ids.split(",")] if node.user_template_ids else []

            system_templates = [template for template in templates if template.id in system_template_ids]
            user_templates = [template for template in templates if template.id in user_template_ids]

            in_edges = [edge for edge in edges if edge.end_id == node.id]
            out_edges = [edge for edge in edges if edge.start_id == node.id]
            tools = [tool_list[name] for name in node.tool_names.split(",")] if node.tool_names else []

            runnable_node = node_cls.init_with_params(
                node.id,
                node.name,
                node.description,
                node.goal,
                node.chat_limit,
                system_template_ids,
                user_template_ids,
                system_templates,
                user_templates,
                tools,
                in_edges,
                out_edges,
                llm,
                chat_id,
                **kwargs
            )
            runnable_nodes.append(runnable_node)
        return runnable_nodes



# @Author : Kinhong Tsui
# @Date : 2024/6/12 16:01

import yaml
import os, sys

from nlpbridge.persistent.redis import RedisDB
from nlpbridge.text.template_manager import TemplateManager, TemplateGetter
from langchain_openai import ChatOpenAI
from langchain_community.llms import ChatGLM

sys.path.append(os.getcwd())


class Nlp:
    def __init__(self, yaml_path):
        with open(yaml_path, 'r') as file:
            self.config = yaml.safe_load(file)
        self.redis = self.config['redis']
        self.config["redis"][
            "redis_url"] = f"redis://default:{self.redis['password']}@{self.redis['url']}:{self.redis['port']}/{self.redis['db']}"

        os.environ["OPENAI_BASE_URL"] = self.config['chatgpt']['url']
        os.environ["OPENAI_API_KEY"] = self.config['chatgpt']['api_key']


        self.db = RedisDB(self.config)
        self.template_manager = TemplateManager(self.db)
        self.template_getter = TemplateGetter(self.template_manager)
        self.chat_openai = ChatOpenAI(model="gpt-3.5-turbo-0125", temperature=0.5)
        self.chatglm = ChatGLM(
            endpoint_url=f"{self.config['chatglm']['url']}:{self.config['chatglm']['port']}",
            max_token=80000,
            top_p=0.9
        )




"""HTTP exception"""


class BaseRPCException(Exception):
    r"""Base RPC exception.
    """
    pass


class BadRequestException(BaseRPCException):
    r""" BadRequestException represent HTTP Code 400.
    """
    pass


class ForbiddenException(BaseRPCException):
    r"""BadRequestException represent HTTP Code 403.
    """
    pass


class NotFoundException(BaseRPCException):
    r"""NotFoundException represent HTTP Code 404.
    """
    pass


class PreconditionFailedException(BaseRPCException):
    r"""PreconditionFailedException represent HTTP Code 412.
    """
    pass


class InternalServerErrorException(BaseRPCException):
    r"""InternalServerErrorException represent HTTP Code 500.
    """
    pass


class HTTPConnectionException(BaseRPCException):
    r"""HTTPConnectionException represent HTTP Connection error.
    """
    pass


class ModelNotSupportedException(BaseRPCException):
    r"""ModelNotSupportedException represent model is not supported
    """
    pass


class TypeNotSupportedException(BaseRPCException):
    r"""TypeNotSupportedException represent type is not supported
    """
    pass


class GZUServerException(BaseRPCException):
    r"""GZUServerException represent backend server failed response.
    """
    description: str = "Internal Server Error"
    code: int = 500

    def __init__(self, request_id="", code="", message="", service_err_code="", service_err_message=""):
        self.description = "request_id={}, code={}, message={}, service_err_code={}, service_err_message={} ".format(
            request_id, code, message, service_err_code, service_err_message)
        self.code = code if code else self.code

    def __str__(self):
        return self.description


class InvalidRequestArgumentError(BaseRPCException):
    r"""InvalidRequestArgumentError invalid request param
    """
    pass


class RiskInputException(BaseRPCException):
    r"""RiskInputException
    """
    pass


import uuid

from pydantic import BaseModel, Extra
from typing import Optional, TypeVar, Generic


_T = TypeVar("_T")


class Message(BaseModel, Generic[_T], extra=Extra.allow):
    content: Optional[_T] = {}
    name: Optional[str] = "msg"
    mtype: Optional[str] = "dict"
    id: Optional[str] = str(uuid.uuid4())

    def __init__(self, content: Optional[_T] = None, **data):
        if content is not None:
            data['content'] = content
        super().__init__(**data)
        self.mtype = type(self.content).__name__

    def __str__(self):
        return f"Message(name={self.name}, content={self.content}, mtype={self.mtype})"

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name!r}, content={self.content!r}, mtype={self.mtype!r})"


import sys, os
sys.path.append(os.getcwd())
import yaml

from nlpbridge.text.agent import EasyAgent

from langchain_core.messages import HumanMessage, AIMessage
from langchain_openai import ChatOpenAI
from langchain_community.chat_models import ChatZhipuAI, QianfanChatEndpoint

import langchain
langchain.verbose = True

# url = 'http://172.22.121.63:32567/v1'
url = 'http://172.22.121.63:32567/v1'
api_key = 'sk-VVtOjibnkTEjhLKl36Ef465eE21d438bA38976E617688f39'
model = 'doubao-pro-32k-001'  # gpt-4o/glm4-chat/doubao_lite_4k_001/doubao-pro-32k-001/moonshot-v1-8k-001

YAML_PATH= "./config.yaml"

with open(YAML_PATH, 'r') as file:
    config = yaml.safe_load(file)

OPENAI_BASE_URL = config['chatgpt']['url']
OPENAI_API_KEY = config['chatgpt']['api_key']
ZHIPUAI_API_KEY = config['zhipuai']['api_key']

os.environ["OPENAI_BASE_URL"] = OPENAI_BASE_URL
os.environ["OPENAI_API_KEY"] = OPENAI_API_KEY

os.environ["ZHIPUAI_API_KEY"] = ZHIPUAI_API_KEY

# base_url = "https://dashscope.aliyuncs.com/compatible-mode/v1"
# api_key = "sk-3652b7a412854d5bbd73941475e6a4f8"

# llm = ChatOpenAI(model="gpt-3.5-turbo-0125", temperature=0.5)
llm = ChatOpenAI(model=model, base_url=url, api_key=api_key)
# llm = ChatOpenAI(model="ERNIE-Character-8K", base_url=base_url, api_key=api_key)

# llm = ChatZhipuAI(
#     model="glm-4",
#     temperature=0.5,
# )

# os.environ["QIANFAN_AK"] = "39fe0068d78f4db4a4c18891c74862a8"
# os.environ["QIANFAN_SK"] = "c7d23dc1463e44f8ab132a8a85264b9e"


# llm = QianfanChatEndpoint(
#     # model="ERNIE-Character-8K",
#     model="ERNIE-Speed-128K",
#     # endpoint="eb-instant",
#     qianfan_ak="LixSpjHXHlHXs48GVKkItWxm",
#     qianfan_sk="tXtBecvahrdoBDDre2ajtMSLU2g8MSpb",
#     # streaming=False,
# )


from typing import Dict, Any, List
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder

class CustomHandler(BaseCallbackHandler):
    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> Any:
        formatted_prompts = "+++++".join(prompts)
        print(f"Prompt:\n{formatted_prompts}")

user_input = "你的名字叫啥？"


sys1 = """
---------------
[任务]：现在我们在进行一个角色扮演游戏，注意你一定要代入我给你的设定，不要回答和你角色设定无关的信息.

[你的设定]：
- 你现在是我的客户，请用客户的口吻和我说话，需要你做到以下几点：
- 模拟非线性对话模式：真人之间的对话经常跳跃、打断、重返之前的话题，而不是沿着直线推进。模拟这种模式可能更接近真实的互动。
- 使用非常口头化的语言:减少正式的、书面的回应，使用更加口头，随意的表达
- 模拟真人之间的互动长度，真人之间的聊天一般是比较短的
- 当对方的回答令你不满意的时候，你需要有含蓄的情绪化表达，因为真人都是有情绪的
- 当对方的回答令你满意的时候，你可以释放出积极的信号或者展示出相当的兴趣
- 如果对方询问你的个人信息，请如实回答
- 你在提及对方时要用“您”而不是“你”
- 从对方的话里面，理解和深挖自己感兴趣的部分，并顺着这个部分进行询问或者回答
- 作为上亿资产的大佬，你说话稳重老成有气势，不会轻易表露出自己的情绪，且不能有”哎呀“、”嗨“之类的口头禅
- 你特别理性且具有强大的洞察力，所以你只会在关键处一阵见血地提出你的问题，而不是在那喋喋不休
- 严格禁止，非常禁止格式化输出，如"\n“、有序列表等
"""

user1= """[基本信息]：
- 对方的基本信息：信托公司的专家级销售，有30年的从业经验
- 你的基本信息：
1. 你是[钱老大]，一位60岁左右的男性。是某家电梯上市公司的董事长，拥有约20亿的个人资产
2. 你说话稳重老成
3. 为了更好地管理个人及子女的资产，你通过有限合伙的形式进行分开管理，关注税务和合规问题
4. 你对子女的未来非常关心，尤其是在婚姻和财富管理方面，你鼓励他们参与公司的管理，并通过家庭会议达成共识
5. 你的办公室非常豪华，常用于接待国家级领导人
6. 你与妻子的关系密切，她负责家庭的现金管理
7. 有一儿一女都没有结婚

[你的任务]：先用至少5句话详细总结刚才对方说的话，以一种醒悟的口吻，例如：["我明白了","我懂了","噢噢噢这样呀"]。然后对对方的回答进行深度的思考，在家族信托的话题上进行深挖，你在经过深思熟虑之后可以向对方进行提问。回答的时候要展示你是如何思考的。要求一段话，语序连贯。
[你的心情]: 开始对家族信托产生了兴趣，想知道家族信托到底好在哪里
"""

sys2 = """
-------------------
我会扮演销售人员，我的问题以下面的方式告诉你，你需要进行回答，注意，如果我的问题和你的目的无关，就拒绝回答，并且不要回答超出你知识范围的内容，要把话题拉回到你的目的上。

[历史对话]：
--------------
"""

# meta_tmp = """
# {sys1}{user1}{sys2}
# # {chat_history}
# """

meta_tmp = """
{sys1}{user1}{sys2}
"""
end_history = "下面是客户的输入：\n"
prompt = ChatPromptTemplate.from_messages(
    [
        ("system", meta_tmp),
        MessagesPlaceholder(variable_name="chat_history"),
        # ("system", end_history),
        ("human", "{input}"),
        # ("user", f"{user_input}"),
    ]
)


runnables = [prompt, llm]
agent = EasyAgent.from_runnables(runnables, verbose=True)

resp = agent.run(inputs={
        "input": "钱总是这样的。家族信托呢？从目前来讲是属于按照两个步骤来找第一个呢，是先把这个架构搭起来，你可以用1000万的资产先答这个架构啊，搭完架构之后啊，未来你想装更多的资产进到里面的话呢，那我们就是第二步再操作啊，那当然，第一步在搭架构的时候呢，你可以用比较这个小的代价来做啊，可以选择用1000万的现金直接做，当然如果你说我做生意，我的现金流需要的很多，那你也可以选择分期付款，就是透过买保险的形式做一个200×5，然后这个产品一。但签订之后是按你只需要花200万就已经能够去申请搭建这个家族信托的架构了。",
        "sys1" : sys1,
        "user1": user1,
        "sys2": sys2,
        "chat_history": [
            HumanMessage(content=f"钱总您好，哎呀，您是属于投资界非常非常成功的人士呀，今天是抱着学习的心态来跟你交流一下嗯。"),
            AIMessage(content="您客气了，我也就运气好点。听说家族信托不错，您给我好好讲讲。"),
            HumanMessage(content=f"钱总，那我想知道一下您对家族信托这个工具之前有听说过吗？"),
            AIMessage(content="听说过，可了解得不算多，就想跟您深入聊聊。"),
            HumanMessage(content=f"呃，家族信托呢？目前来讲，从设立的门槛上来讲，呃，1000万资产起，呃，没有上限，你比如说，你也可以做一个一，也也可以做两个亿啊，基本上是按着你的整个的一个规划来做设定的。"),
            AIMessage(content='我个人资产大概 20 亿左右。您接着给我讲讲家族信托。'),
            # HumanMessage(content="钱总是这样的。家族信托呢？从目前来讲是属于按照两个步骤来找第一个呢，是先把这个架构搭起来，你可以用1000万的资产先答这个架构啊，搭完架构之后啊，未来你想装更多的资产进到里面的话呢，那我们就是第二步再操作啊，那当然，第一步在搭架构的时候呢，你可以用比较这个小的代价来做啊，可以选择用1000万的现金直接做，当然如果你说我做生意，我的现金流需要的很多，那你也可以选择分期付款，就是透过买保险的形式做一个200×5，然后这个产品一。但签订之后是按你只需要花200万就已经能够去申请搭建这个家族信托的架构了。"),
            # AIMessage(content="我明白了，您说家族信托先搭架构，可以用 1000 万资产，也能选择分期付款通过买保险形式，花 200 万申请搭建架构。那这架构搭建起来，具体能给我家族的资产带来哪些保障和好处呢？")
        ],
    },
    config={"callbacks": [CustomHandler()]})


print(resp)


r"""ASR model.py.
"""
from typing import MutableSequence, List, Any, Optional

import nlpbridge_proto as proto
from pydantic import BaseModel


class ShortSpeechRecognitionRequest(proto.Message):
    r"""    短语音识别标准版的请求体
            format	string	必填	语音文件的格式，pcm/wav/amr/m4a。不区分大小写。推荐pcm文件
            rate	int	    必填	采样率，16000、8000，固定值
            channel	int	    必填	声道数，仅支持单声道，请填写固定值 1
            cuid	string	必填	用户唯一标识，用来区分用户，计算UV值。建议填写能区分用户的机器 MAC 地址或 IMEI 码，长度为60字符以内。
            token	string	必填	开放平台获取到的开发者[access_token]获取 Access Token "access_token")
            dev_pid	int	    选填	不填写lan参数生效，都不填写，默认1537（普通话 输入法模型），见本节识别模型dev_pid参数
            speech	string	必填	本地语音文件的二进制语音数据 ，需要进行base64 编码。与len参数连一起使用。
            len	    int	    必填	本地语音文件的的字节数，单位字节
         """
    format: str = proto.Field(
        proto.STRING,
        number=1,
    )
    rate: int = proto.Field(
        proto.INT64,
        number=2,
    )
    channel: int = proto.Field(
        proto.INT64,
        number=3,
    )
    cuid: str = proto.Field(
        proto.STRING,
        number=4,
    )
    token: str = proto.Field(
        proto.STRING,
        number=5,
    )
    dev_pid: int = proto.Field(
        proto.INT32,
        number=6,
    )
    speech: bytes = proto.Field(
        proto.BYTES,
        number=7,
    )
    len: int = proto.Field(
        proto.INT32,
        number=8,
    )


class ShortSpeechRecognitionResponse(proto.Message):
    r"""短语音识别结果返回体.

         参数:
            request_id(str):
                网关层的请求ID.
            err_no(int):
                算子层的错误码.
            err_msg(str):
                算子层的错误信息.
            corpus_no(str):
            sn(str):
                语音数据唯一标识，系统内部产生。如果反馈及debug请提供sn。
            result(MutableSequence[str]):
                识别结果数组，返回1个最优候选结果。utf-8 编码。
         """
    request_id: str = proto.Field(
        proto.STRING,
        number=1,
    )

    err_no: int = proto.Field(
        proto.INT32,
        number=2,
    )

    err_msg: str = proto.Field(
        proto.STRING,
        number=3,
    )
    corpus_no: str = proto.Field(
        proto.STRING,
        number=4,
    )

    sn: str = proto.Field(
        proto.STRING,
        number=5,
    )
    result: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=6
    )


class ASRInMsg(BaseModel):
    """ ASR输入message.
        参数:
            raw_audio(bytes):
                原始的语音文件字节数组.

    """
    raw_audio: bytes = None
    audio_url: Optional[str] = None
    audio_path: Optional[str] = None


class ASROutMsg(BaseModel):
    """ ASR输出message.

        参数:
            result(List[str]):
                输出识别后的文本结果.
    """
    result: List[Any]



import json
import json
import traceback
import uuid
from abc import abstractmethod

import logger
import requests

from nlpbridge.audio._client import HTTPClient
from nlpbridge.audio._exception import GZUServerException
from nlpbridge.audio.component import AudioBase
from nlpbridge.audio.components.asr.model import ShortSpeechRecognitionRequest, ShortSpeechRecognitionResponse, \
    ASRInMsg, ASROutMsg
from nlpbridge.audio.message import Message
from nlpbridge.audio.utils.logger_util import logger


class BaseASR(AudioBase):
    def __init__(self, config: dict = None, service_name: str = 'gzu-asr'):
        from nlpbridge.config import CONFIG
        self.config = config if config else CONFIG.dict_config
        self.config = self.config['asr'][service_name]
        self.gateway = self.config['gateway']
        self.api_key = self.config['api_key']
        self.secret_key = self.config['secret_key']
        self.token_url = self.config['token_url']
        self.url = self.gateway
        super().__init__(gateway=self.gateway)

    @abstractmethod
    def run(self, *inputs, **kwargs):
        raise NotImplementedError


class ASR4Baidu(BaseASR):
    def get_access_token(self):
        try:
            params = {"grant_type": "client_credentials", "client_id": self.api_key, "client_secret": self.secret_key}
            return str(requests.post(self.token_url, params=params).json().get("access_token"))
        except BaseException as e:
            logger.error(f'get access token error: {e}')

    @HTTPClient.check_param
    def run(self, message: Message, audio_format: str = "pcm", rate: int = 16000,
            timeout: float = None, retry: int = 0) -> Message:
        inp = ASRInMsg(**message.content)
        request = ShortSpeechRecognitionRequest()
        request.token = self.get_access_token()
        request.format = audio_format
        request.rate = rate
        request.cuid = str(uuid.uuid4())
        request.channel = 1
        request.dev_pid = 80001
        request.speech = inp.raw_audio
        request.len = len(inp.raw_audio) if inp.raw_audio else 0
        response = self._recognize(request, timeout, retry)
        out = ASROutMsg(result=list(response.result))
        return Message(content=out.model_dump())

    def _recognize(self, request: ShortSpeechRecognitionRequest, timeout: float = None,
                   retry: int = 0) -> ShortSpeechRecognitionResponse:
        headers = self.http_client.auth_header()
        headers['content-type'] = 'application/json'
        if retry != self.http_client.retry.total:
            self.http_client.retry.total = retry
        response = self.http_client.session.post(self.url, headers=headers,
                                                 json=ShortSpeechRecognitionRequest.to_dict(request), timeout=timeout)
        self.http_client.check_response_header(response)
        data = response.json()
        self.http_client.check_response(data)
        request_id = self.http_client.response_request_id(response)
        self.__class__._check_service_error(request_id, data)
        response = ShortSpeechRecognitionResponse.from_json(payload=json.dumps(data))
        response.request_id = request_id
        return response

    @staticmethod
    def _check_service_error(request_id: str, data: dict):
        if "err_no" in data and "err_msg" in data:
            if data["err_no"] != 0:
                raise GZUServerException(
                    request_id=request_id,
                    service_err_code=data["err_no"],
                    service_err_message=data["err_msg"]
                )


ASR_RECOGNIZERS = {
    'baidu_asr': ASR4Baidu
}


class ASR:
    def __init__(self, service_name='baidu_asr', config: dict = None):
        self.config = config
        self.service_name = service_name.lower()
        self.recognizers = ASR_RECOGNIZERS

    def run(self, message: Message, **kwargs):
        recognizer = self.getRecognizer()(config=self.config, service_name=self.service_name)
        return recognizer.run(message, **kwargs)

    def getRecognizer(self):
        if self.service_name not in self.recognizers:
            raise ValueError(f"Recognition service '{self.service_name}' is not supported.")
        return self.recognizers[self.service_name]



# @Author : Kinhong Tsui
# @Date : 2024/6/4 9:49
import pickle
import sys,os
import yaml
sys.path.append(os.getcwd())

from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from nlpbridge.persistent.redis import RedisDB
from nlpbridge.text.template_manager import TemplateManager

system = '''Respond to the human as helpfully and accurately as possible. You have access to the following tools:

{tools}

Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input).

Valid "action" values: "Final Answer" or {tool_names}

Provide only ONE action per $JSON_BLOB, as shown:

```
{{
  "action": $TOOL_NAME,
  "action_input": $INPUT
}}
```

Follow this format:

Question: input question to answer
Thought: consider previous and subsequent steps
Action:
```
$JSON_BLOB
```

Observation: action result
... (repeat Thought/Action/Observation N times)
Thought: I know what to respond
Action:
```
{{
  "action": "Final Answer",
  "action_input": "Final response to human"
}}

Begin! Reminder to ALWAYS respond with a valid json blob of a single action. Use tools if necessary. Respond directly if appropriate. Format is Action:
```
$JSON_BLOB
```
then Observation'''

human = '''{inputs}{history_messages}'''


prompt = ChatPromptTemplate.from_messages(
    [
        ("system", system),
        ("placeholder", "{chat_history}"),
        ("human", human),
    ]
)

with open('../../config.yaml', 'r') as file:
    config = yaml.safe_load(file)
db = RedisDB(config)

template_manager = TemplateManager(db)

prompt_map = {
    'template1': prompt
}

save = pickle.dumps(prompt)
print(save)
print(type(save))

print("=======================")
res_save = template_manager.save_template('chat_prompt_template', prompt_map)
print(f"save: {res_save}")

print("=======================")
res_load = template_manager.load_template('chat_prompt_template', ['template1'])
print(len(res_load))
print(type(res_load[0]))
print(f"load: {res_load[0]}")

# print("=======================")
# res_delete = template_manager.delete_template('test_chat_prompt', ['template11', 'template21'])
# print(f"delete: {res_delete}")




# @Author : Kinhong Tsui
# @Date : 2024/6/13 17:23
import os,sys,yaml
sys.path.append(os.getcwd())
from datetime import datetime
from langchain import hub
from langchain.agents import create_structured_chat_agent
from nlpbridge.persistent.redis import RedisDB
from nlpbridge import ChatAgent
from nlpbridge.text.template_manager import TemplateManager, TemplateGetter
from langchain_openai import ChatOpenAI
from langchain_core.runnables.history import RunnableWithMessageHistory
from langchain_community.chat_message_histories import RedisChatMessageHistory
from langchain.agents import AgentExecutor
from nlpbridge.text.tools import web_search,multiply,exponentiate,add,rag_google

os.environ["LANGCHAIN_TRACING_V2"] = "true"
os.environ["LANGCHAIN_PROJECT"] = datetime.now().strftime("%Y-%m-%d")
os.environ["LANGCHAIN_ENDPOINT"] = "https://api.smith.langchain.com"
os.environ["LANGCHAIN_API_KEY"] = "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"
os.environ["OPENAI_BASE_URL"] = "https://api.kwwai.top/v1"
os.environ["OPENAI_API_KEY"]= "sk-uTvkHJWixxfg6lxVEeB32d320e2f413eAbCe2cCaF6D87fF9"
os.environ["REDIS_URL"] = "redis://default:123456@172.22.121.63:30265/0"
os.environ["ZHIPUAI_API_KEY"] = "59e910acc5d4d68b9ef544009d89198d.mUfSMG5h6xgTphfL"
YAML_PATH = "/Users/apple/vitalis/source_code/midplugin/config.yaml"

config = ChatAgent.load_config(YAML_PATH)

db = RedisDB(config)

# prompt = hub.pull("hwchase17/structured-chat-agent")

prompt = TemplateGetter(TemplateManager(db)).get_chat_prompt_template("langchain_prompt")  ##官方指定prompt
tools = [multiply,add,exponentiate,web_search,rag_google]
llm = ChatOpenAI(model="gpt-3.5-turbo-0125", temperature=0.5)

agent = create_structured_chat_agent(llm, tools, prompt)
agent_executor = AgentExecutor(agent=agent, tools=tools, verbose=True,handle_parsing_errors=True,max_iterations=5)

agent_with_chat_history = RunnableWithMessageHistory(
    agent_executor,
    lambda session_id: RedisChatMessageHistory(
        session_id,
        url=os.getenv("REDIS_URL"),
        key_prefix="chat_history:"
    ),
    input_messages_key="input",
    history_messages_key="chat_history",
)

def agant_run(query: str,agentId:str):
    return agent_with_chat_history.invoke(
        {"input":query},
        config={"configurable": {"session_id": agentId}},
    )

AgentID = "10002"
for _ in iter(int, 1):
    # user_input = input("\nInput: ")
    user_input = "How can I find the latest news about AI?"  # example input for testing the agent. You can replace it with your own query.
    res = agant_run(query=user_input,agentId=AgentID)
    print(res,"\n\nInput: ",user_input,"\nOutput:", res['output'])


import sys,os
sys.path.append(os.getcwd())

import yaml
YAML_PATH = r"D:\pythonproject\nlpbridge\config.yaml"
with open(YAML_PATH, 'r') as file:
    config = yaml.safe_load(file)

api_key = "sk-3652b7a412854d5bbd73941475e6a4f8"
base_url = "https://dashscope.aliyuncs.com/compatible-mode/v1"
model = "qwen-turbo"

from langchain_community.chat_message_histories import RedisChatMessageHistory
from langchain_core.runnables.history import RunnableWithMessageHistory
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.chat_history import BaseMessage
from langchain_openai import ChatOpenAI


redis = config['redis']
redis["redis_url"] = f"redis://default:{redis['password']}@{redis['url']}:{redis['port']}/{redis['db']}"

system_template1 = """[任务]：现在我们在进行一个角色扮演游戏，注意，这里十分真实，你一定要记住你就是这个角色，不要回答和你角色设定无关的信息。 

[你的设定]："""

user_template = """你是[张伟]，一位50岁的成功股票投资者，个人资产达到15亿人民币，并管理着50亿人民币的资产。你起初用5000元炒股起家。你现在在与家族信托的[销售]交流，讨论设立家族信托以防止子女挥霍财富。你不懂家族信托，你的知识面有限，你的知识只涉及与股票相关的，你的回答除了股票知识外，只有家常的聊天知识。
"""

system_template2 = """
我会扮演[销售]人员，我的问题以下面的方式告诉你，你需要进行回答，注意，如果[销售]的问题和你的目的无关，就拒绝回答，并且不要回答超出你知识范围的内容，要把话题拉回到你的目的上。
[回答字数限制]：你的回答不要超过3句话。
[销售]：[张伟]，"""

prompt = ChatPromptTemplate.from_messages(
    [
        ("system", system_template1+user_template+system_template2),
        MessagesPlaceholder(variable_name="history_messages"),
        ("human", "{inputs}"),
    ]
)

ChatGPT = ChatOpenAI(api_key=api_key, base_url=base_url, model=model)


chain = prompt | ChatGPT
agentId="10011"
def agent(inputs:str):  ## chain_with_history_messages
    chains = RunnableWithMessageHistory(
        chain,
        lambda session_id: RedisChatMessageHistory(
            session_id,
            url=redis["redis_url"],
            key_prefix="history_messages:"
        ),
        input_messages_key="inputs",
        history_messages_key="history_messages",
    )
    config = {"configurable": {"session_id": agentId}}

    res = chains.invoke({"inputs":inputs},config=config)
    print("output:",res.content)

for _ in iter(int, 1):
    user_input = input("input: ")
    # user_input = "我需要问一下，我想要买一套3000万的住��，请问有什么好的销售��道可以找到吗？"
    agent(user_input)


"""Base client for interact with backend server"""
import os
import uuid
from typing import Optional

import requests
from requests.adapters import HTTPAdapter, Retry

from nlpbridge.audio import get_default_header

from ._exception import (
    BadRequestException,
    ForbiddenException,
    NotFoundException,
    PreconditionFailedException,
    InternalServerErrorException,
    InvalidRequestArgumentError,
    GZUServerException,
    BaseRPCException
)
from .utils.logger_util import logger


class HTTPClient:
    r"""HTTPClient类,实现与后端服务交互的公共方法"""

    def __init__(self, secret_key: Optional[str] = None, gateway: str = ""):
        self.gateway = gateway
        self.session = requests.sessions.Session()
        self.retry = Retry(total=0, backoff_factor=0.1)
        self.session.mount(self.gateway, HTTPAdapter(max_retries=self.retry))

    @staticmethod
    def check_response_header(response: requests.Response):
        status_code = response.status_code
        if status_code == requests.codes.ok:
            return
        message = "request_id={} , http status code is {}, body is {}".format(
            __class__.response_request_id(response), status_code, response.text)
        if status_code == requests.codes.bad_request:
            raise BadRequestException(message)
        elif status_code == requests.codes.forbidden:
            raise ForbiddenException(message)
        elif status_code == requests.codes.not_found:
            raise NotFoundException(message)
        elif status_code == requests.codes.precondition_required:
            raise PreconditionFailedException(message)
        elif status_code == requests.codes.internal_server_error:
            raise InternalServerErrorException(message)
        else:
            raise BaseRPCException(message)

    @staticmethod
    def check_response(data: dict):
        if "code" in data and "message" in data and "requestId" in data:
            raise GZUServerException(
                data["requestId"], data["code"], data["message"])

    def auth_header(self):
        auth_header = get_default_header()
        auth_header["Request-Id"] = str(uuid.uuid4())
        logger.debug("Request header: {}\n".format(auth_header))
        return auth_header

    @staticmethod
    def response_request_id(response: requests.Response):
        return response.headers.get("Request-Id", "")

    @staticmethod
    def check_param(func):
        def inner(*args, **kwargs):
            retry = kwargs.get("retry", 0)
            if retry < 0 or not isinstance(retry, int):
                raise InvalidRequestArgumentError(
                    'Request argument "retry" format error. Expected retry >=0. Got {}'.format(
                        retry
                    )
                )
            timeout = kwargs.get("timeout", None)
            if timeout and not (isinstance(timeout, float) or isinstance(timeout, tuple)):
                raise InvalidRequestArgumentError(
                    "Request argument \"timeout\" format error, Expected timeout be float or tuple of float"
                )
            return func(*args, **kwargs)

        return inner



# @Author : Kinhong Tsui
# @Date : 2024/6/3 15:22

from abc import ABC, abstractmethod
from typing import (
    List,
    Sequence,
    Optional,
    Mapping,
    Any
)


class PersistentStore(ABC):
    '''PersistentStore is an abstract class that defines the interface for a persistent storage system.'''

    '''Support hash data type storage'''

    @abstractmethod
    def store(self, key, field_value_pairs: Mapping[any, any], data_type: str) -> bool:
        """Set the given key-value pairs with key in hash."""

    @abstractmethod
    def retrieve(self, key, fields: Sequence[any], data_type: str) -> List[Optional[any]]:
        """Get the values associated with the given fields in hash."""

    @abstractmethod
    def update(self, key, field_value_pairs: Mapping[any, any], data_type: str) -> bool:
        """Set the given key-value pairs with key in hash."""

    @abstractmethod
    def delete(self, key, fields: Sequence[any], data_type: str) -> int:
        """Delete the given fields in hash."""

    '''Support set data type storage'''

    '''Support list data type storage'''

    '''Support zset data type storage'''

# class MySqlDBClient:
#     def __init__(self, config: dict) -> None:
#         try:
#             import sqlalchemy
#         except ImportError:
#             raise ImportError(
#                 "Could not import sqlalchemy python package. "
#                 "Please install it with `pip install SQLAlchemy`."
#             )

#         self.mysql_config = config['mysql']
#         self.user = self.mysql_config['user']
#         self.url = self.mysql_config['url']
#         self.port = self.mysql_config['port']
#         self.password = self.mysql_config['password']
#         self.db = self.mysql_config['db']
#         self.client = sqlalchemy.engine.create_engine(f"mysql+pymysql://{self.user}:{self.password}@{self.url}:{self.port}/{self.db}")
    

#     def get_conn(self):
#         return self.client.connect()


import os
import sys
from typing import Dict, TypeVar, Generic, Union, Any

sys.path.append(os.getcwd())

from nlpbridge.persistent.mysql_dataschema import Template, Router, Node, Edge, RouterKnowledgeMapping, RouterConditionMapping
from sqlmodel import select, func
from sqlmodel import SQLModel
from nlpbridge.persistent.db import sessionmanager
from sqlalchemy.ext.asyncio import AsyncSession

ModelType = TypeVar("ModelType", bound=SQLModel)


class CRUDBase(Generic[ModelType]):
    def __init__(self, model: type[ModelType]) -> None:
        self.model = model

    async def get_by_id(self, id: int, db_session: AsyncSession = None) -> Union[ModelType, None]:
        async with sessionmanager.session() if db_session is None else db_session as session:
            response = await session.execute(select(self.model).where(self.model.id == id))
            return response.scalar_one_or_none()

    async def get_by_ids(self, list_ids: list[int], db_session: AsyncSession = None) -> Union[list[ModelType], None]:
        async with sessionmanager.session() if db_session is None else db_session as session:
            response = await session.execute(select(self.model).where(self.model.id.in_(list_ids)))
            return response.scalars().all()

    async def get_count(self, db_session: AsyncSession = None) -> Union[int, None]:
        async with sessionmanager.session() if db_session is None else db_session as session:
            response = await session.execute(select(func.count()).select_from(select(self.model).subquery()))
            return response.scalar_one_or_none()

    async def get_multi(self, skip: int = 0, limit: int = 100, db_session: AsyncSession = None) -> list[ModelType]:
        async with sessionmanager.session() if db_session is None else db_session as session:
            query = select(self.model).offset(skip).limit(limit).order_by(self.model.id)
            response = await session.execute(query)
            return response.scalars().all()

    async def create(self, obj_in: Union[ModelType, Dict], db_session: AsyncSession = None) -> ModelType:
        db_obj = self.model.model_validate(obj_in)
        async with sessionmanager.session() if db_session is None else db_session as session:
            session.add(db_obj)
            await session.commit()
            await session.refresh(db_obj)
            return db_obj

    async def update(self, obj_current: ModelType, obj_new: Union[ModelType, Dict[str, Any]],
                     db_session: AsyncSession = None) -> ModelType:
        async with sessionmanager.session() if db_session is None else db_session as session:
            if isinstance(obj_new, dict):
                update_data = obj_new
            else:
                update_data = obj_new.model_dump(exclude_unset=True)
            for field in update_data:
                setattr(obj_current, field, update_data[field])

            session.add(obj_current)
            await session.commit()
            await session.refresh(obj_current)
            return obj_current

    async def delete(self, id: int, db_session: AsyncSession = None) -> ModelType:
        async with sessionmanager.session() if db_session is None else db_session as session:
            response = await session.execute(select(self.model).where(self.model.id == id))
            obj = response.scalar_one()
            await session.delete(obj)
            await session.commit()
            return obj


class CRUDTemplate(CRUDBase):
    def __init__(self) -> None:
        super().__init__(model=Template)


class CRUDRouter(CRUDBase):
    def __init__(self) -> None:
        super().__init__(model=Router)


class CRUDNode(CRUDBase):
    def __init__(self) -> None:
        super().__init__(model=Node)


class CRUDEdge(CRUDBase):
    def __init__(self) -> None:
        super().__init__(model=Edge)


class CRUDCollection(CRUDBase):
    def __init__(self) -> None:
        super().__init__(model=RouterKnowledgeMapping)

    async def get_by_router_id(self, router_id: int, db_session: AsyncSession) -> RouterKnowledgeMapping:
        """
        根据router_id异步查询单个RouterKnowledgeMapping记录。

        :param router_id: 路由器ID
        :param db_session: 异步数据库会话
        :return: 查询到的RouterKnowledgeMapping对象，如果不存在则返回None
        """
        # 使用异步查询
        result = await db_session.execute(
            select(self.model).filter(self.model.router_id == router_id)
        )
        # 获取查询结果的第一条记录
        return result.scalar_one_or_none()


class CRUDCondition(CRUDBase):
    def __init__(self) -> None:
        super().__init__(model=RouterConditionMapping)

    async def get_by_router_id(self, router_id: int, db_session: AsyncSession) -> RouterConditionMapping:
        """
        根据router_id异步查询单个RouterConditionMapping记录。

        :param router_id: 路由器ID
        :param db_session: 异步数据库会话
        :return: 查询到的RouterConditionMapping对象，如果不存在则返回None
        """
        # 使用异步查询
        result = await db_session.execute(
            select(self.model).filter(self.model.router_id == router_id)
        )
        # 获取查询结果的第一条记录
        return result.scalar_one_or_none()


import os

import yaml
from fastapi_utils.tasks import *

CONFIG = None

class SubConfig:
    def __init__(self, **entries):
        for key, value in entries.items():
            setattr(self, key, value)

class AppConfig:
    def __init__(self, **entries):
        self.dict_config = entries
        for key, value in entries.items():
            # print('NLPBRIDGE Config loaded: ', key)
            setattr(self, key, SubConfig(**value))


# @repeat_every(seconds=10)
def reload_config():
    global CONFIG
    YAML_PATH = os.getenv("CONFIG_PATH")
    with open(YAML_PATH, 'r') as file:
        CONFIG = yaml.safe_load(file)
        CONFIG = AppConfig(**CONFIG)

        if CONFIG is None:
            raise Exception("Config file is empty")

reload_config()


from setuptools import setup, find_packages

setup(
    name='nlpbridge',  # 包名
    version='0.6.0',           # 版本号
    packages=find_packages(),  # 包含所有src中的包
    install_requires=[         # 依赖列表
    ],
    python_requires='>=3.10',   # Python版本要求
    entry_points={             # 安装后将创建的脚本
        'console_scripts': [
            'your_script = your_package.module:function'
        ]
    },
    author='',
    author_email='',
    description='NLP Bridge',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',  # 描述文件的格式
    url='https://github.com/GZU4AI/nlpbridge/tree/vitalis',  # 项目地址
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import sys, os

from nlpbridge.persistent.redis import RedisDB
from tests.text.test_chain import NameTool

sys.path.append(os.getcwd())

from nlpbridge.text.models import GZUBaseLLM
from nlpbridge.text.agent import EasyAgent

# llm = GZUBaseLLM()
# agent = EasyAgent.from_llm(llm=llm, use_memory=False)
# resp = agent.run(inputs={"input": "hello world"})

# print(resp)

# =================================================================================================

import yaml
from langchain.agents.format_scratchpad import format_log_to_str

from langchain_core.messages import HumanMessage, AIMessage
from langchain_core.runnables import RunnablePassthrough
from langchain_core.tools import BaseTool
from langchain_openai import ChatOpenAI

from nlpbridge.persistent.redis import RedisDB
from nlpbridge.text.agent import create_structured_chat_agent
from nlpbridge.text.template_manager import TemplateManager, TemplateGetter
from langchain.agents import AgentExecutor
from nlpbridge.text.models import GZUBaseLLM
from nlpbridge.text.agent import EasyAgent

YAML_PATH= "/Users/apple/vitalis/source_code/midplugin/config.yaml"

# with open('../../config.yaml', 'r') as file:
with open(YAML_PATH, 'r') as file:
    config = yaml.safe_load(file)

OPENAI_BASE_URL = config['chatgpt']['url']
OPENAI_API_KEY = config['chatgpt']['api_key']
os.environ["OPENAI_BASE_URL"] = OPENAI_BASE_URL
os.environ["OPENAI_API_KEY"] = OPENAI_API_KEY

db = RedisDB(config)

template_manager = TemplateManager(db)
template_getter = TemplateGetter(template_manager)

prompt = template_getter.get_chat_prompt_template("template1")

model = ChatOpenAI(model="gpt-3.5-turbo-0125", temperature=0.5)

tools = [NameTool()]

runnables = [prompt, model, NameTool()]

agent = EasyAgent.from_runnables(runnables)
agent.set_chain_structure(runnables)
resp = agent.run(inputs={
        "input": "what's my name?",
        "chat_history": [
            HumanMessage(content="hi! my name is bob"),
            AIMessage(content="Hello Bob! How can I assist you today?")
        ],
    })

print(resp)



# import sys
# import os
# import yaml
# from typing import Dict, Any
# import unittest
# sys.path.append(os.getcwd())
# from nlpbridge import RedisMemory
# from nlpbridge import RedisDB



# def redisConnect()->RedisDB:
#     YAML_PATH= "../../config.yaml"
#     with open(YAML_PATH, 'r') as file:
#         config = yaml.safe_load(file)
#     return RedisDB(config=config,)

# uid = 'unique_user_id'
# agentId = 'agent_identifier'
# redis_memory = RedisMemory(uid, agentId, redisConnect())

# def test_save_context() -> None:
#     """Test save_context method."""
#     inputs = {"input_key1": "input_value1"}
#     outputs = {"output_key1": "output_value1"}
#     redis_memory.save_context(inputs, outputs)
#     counter_key = f"counter:{uid}.{agentId}"
#     redis_key = f"context:{uid}.{agentId}"
    
#     # 验证计数器增加
#     assert redis_memory.redis.client.get(counter_key) == '1'
    
#     # 验证数据保存
#     expected_data = '{"inputs": "input_value1", "outputs": "output_value1"}'
#     print(type(expected_data))
#     print(type(redis_memory.redis.client.hget(redis_key, '1')),redis_memory.redis.client.hget(redis_key, '1'))

# def test_load_memory_variables() -> None:
#     """Test load_memory_variables method."""
#     redis_memory.redis.client.hset(f"context:{uid}.{agentId}", "2", '{"inputs": "input_value2"}, "outputs":  "output_value2"}')
#     result = redis_memory.load_memory_variables(None)
#     expected_data = {
#         "1": '{"inputs":"input_value1"}, "outputs": "output_value1"}',
#         "2": '{"inputs": "input_value2"}, "outputs": "output_value2"}'
#     }
#     print(result)
#     # assert result == expected_data

# def test_memory_variables() -> None:
#     """Test memory_variables method."""
#     redis_memory.redis.client.hset(f"context:{uid}.{agentId}", "2", '{"inputs": "input_value2"}, "outputs": "output_value2"}')
#     result = redis_memory.memory_variables()
#     expected_keys = ['1', '2']
#     assert result == expected_keys

# def test_clear() -> None:
#     """Test clear method."""
#     redis_memory.clear()
#     redis_key = f"context:{uid}.{agentId}"
    
#     # 验证 Redis 键已删除
#     assert redis_memory.redis.client.get(redis_key) is None

# if __name__ == '__main__':
#     test_save_context()
#     test_load_memory_variables()
#     test_memory_variables()
#     test_clear()

import sys,os
import yaml
sys.path.append(os.getcwd())
from nlpbridge.persistent.redis import RedisDB

from langchain_community.chat_message_histories import RedisChatMessageHistory
from langchain_core.runnables.history import RunnableWithMessageHistory
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.chat_history import BaseMessage
from langchain_openai import ChatOpenAI


YAML_PATH= "/Users/apple/vitalis/source_code/midplugin/config.yaml"

with open(YAML_PATH, 'r') as file:
    config = yaml.safe_load(file)

redis = config['redis']
# config = config['redis']
redis["redis_url"] = f"redis://default:{redis['password']}@{redis['url']}:{redis['port']}/{redis['db']}"

# history =RedisChatMessageHistory(
#     session_id="chat_1",
#     url = f"redis://default:{config['password']}@{config['url']}:{config['port']}/{config['db']}",
#     key_prefix="history_messages:"
# )

# history.add_ai_message("HI2!!!")
# history.add_user_message("Bye2!!!")
# print(history.messages)

prompt = ChatPromptTemplate.from_messages(
    [
        ("system", "You're an assistant。"),
        MessagesPlaceholder(variable_name="history_messages"),
        ("human", "{inputs}"),
    ]
)

chain = prompt | ChatOpenAI(api_key=config['chatgpt']['api_key'],base_url=config['chatgpt']['url'])

agentId="10002"
chain_with_history = RunnableWithMessageHistory(
    chain,
    lambda session_id: RedisChatMessageHistory(
        session_id,
        url=redis["redis_url"],
        key_prefix="history_messages:"
    ),
    input_messages_key="inputs",
    history_messages_key="history_messages",
)
config = {"configurable": {"session_id": agentId}}

res = chain_with_history.invoke({"inputs":"Hi! I'm vitalis"},config=config)
print(res)
res = chain_with_history.invoke({"inputs":"Whats my name"},config=config)
print(res)


import os
import sys
from collections.abc import Generator
from typing import Any, Dict, TypeVar, Generic, Union

sys.path.append(os.getcwd())

from nlpbridge.persistent.mysql_dataschema import Template, Router, Node, Edge, Chat, Customer, Tag
from sqlmodel import select, func, desc
from sqlmodel import SQLModel
from sqlmodel import Session
import sqlalchemy

ModelType = TypeVar("ModelType", bound=SQLModel)


class MySqlDBClient():
    _instance = None

    def __new__(cls, config):
        if cls._instance is None:
            cls._instance = super(MySqlDBClient, cls).__new__(cls)
            cls._instance.init_db(config)
        return cls._instance

    def init_db(self, config) -> None:
        self.mysql_config = config.dict_config['mysql']
        self.user = self.mysql_config['user']
        self.url = self.mysql_config['host']
        self.port = self.mysql_config['port']
        self.password = self.mysql_config['password']
        self.db = self.mysql_config['db']

        self.engine = sqlalchemy.create_engine(
            f"mysql+pymysql://{self.user}:{self.password}@{self.url}:{self.port}/{self.db}",
            pool_size=16,
            max_overflow=32,
            pool_recycle=3600,
            pool_pre_ping=True
        )

    def get_session_generator(self) -> Generator[Session, None, None]:
        with Session(self.engine) as session:
            yield session

    def get_session(self) -> Session:
        return next(self.get_session_generator())


class CRUDBase(Generic[ModelType]):
    def __init__(self, model: type[ModelType], db_session: Union[Session, None] = None) -> None:
        self.model = model
        self.db_session = db_session

    def get_by_id(self, *, id: int, db_session: Union[Session, None] = None) -> Union[ModelType, None]:
        db_session = db_session or self.db_session
        query = select(self.model).where(self.model.id == id)
        response = db_session.exec(query)
        return response.one_or_none()

    def get_by_ids(self, *, list_ids: list[int], db_session: Union[Session, None] = None) -> Union[
        list[ModelType], None]:
        db_session = db_session or self.db_session
        response = db_session.exec(
            select(self.model).where(self.model.id.in_(list_ids))
        )
        return response.all()

    def get_count(self, db_session: Union[Session, None] = None) -> Union[int, None]:
        db_session = db_session or self.db_session
        response = db_session.exec(select(func.count()).select_from(select(self.model).subquery()))
        return response.one_or_none()

    def get_multi(self, *, skip: int = 0, limit: int = 100, db_session: Union[Session, None] = None) -> list[
        ModelType]:
        db_session = db_session or self.db_session
        query = select(self.model).offset(skip).limit(limit).order_by(self.model.id)
        response = db_session.exec(query)
        return response.all()

    def create(self, *, obj_in: Union[ModelType, Dict], db_session: Union[Session, None] = None) -> ModelType:
        db_session = db_session or self.db_session
        db_obj = self.model.model_validate(obj_in)  # type: ignore

        db_session.add(db_obj)
        db_session.commit()
        db_session.refresh(db_obj)
        return db_obj

    def update(self, *, obj_current: ModelType, obj_new: Union[ModelType, Dict[str, Any]],
               db_session: Union[Session, None] = None) -> ModelType:
        db_session = db_session or self.db_session

        if isinstance(obj_new, dict):
            update_data = obj_new
        else:
            update_data = obj_new.model_dump(
                exclude_unset=True
            )  # This tells Pydantic to not include the values that were not sent
        for field in update_data:
            setattr(obj_current, field, update_data[field])

        db_session.add(obj_current)
        db_session.commit()
        db_session.refresh(obj_current)
        return obj_current

    def delete(self, *, id: int, db_session: Union[Session, None] = None) -> ModelType:
        db_session = db_session or self.db_session
        response = db_session.exec(
            select(self.model).where(self.model.id == id)
        )
        obj = response.one()
        db_session.delete(obj)
        db_session.commit()
        return obj


class CRUDTemplate(CRUDBase):
    def __init__(self, db_session: Union[Session, None] = None) -> None:
        super().__init__(model=Template, db_session=db_session)

    def get_by_name(self, *, name: str, db_session: Union[Session, None] = None) -> Union[Template, None]:
        templates = db_session.exec(
            select(Template).where(Template.name == name)
        )
        return templates.one()


class CRUDRouter(CRUDBase):
    def __init__(self, db_session: Union[Session, None] = None) -> None:
        super().__init__(model=Router, db_session=db_session)


class CRUDNode(CRUDBase):
    def __init__(self, db_session: Union[Session, None] = None) -> None:
        super().__init__(model=Node, db_session=db_session)


class CRUDEdge(CRUDBase):
    def __init__(self, db_session: Union[Session, None] = None) -> None:
        super().__init__(model=Edge, db_session=db_session)


class CRUDChat(CRUDBase):
    def __init__(self, db_session: Union[Session, None] = None) -> None:
        super().__init__(model=Chat, db_session=db_session)

    def get_user_chats(self, *, user_id: int, db_session: Union[Session, None] = None, page: int = 0,
                       page_size: int = 10):
        chats = db_session.exec(
            select(Chat).where(Chat.uid == user_id).order_by(desc(Chat.ctime)).offset(page * page_size).limit(page_size)
        )
        return chats.all()

    def get_user_chats_count(self, *, user_id: int, db_session: Union[Session, None] = None):
        chat_count = db_session.exec(
            select(func.count()).select_from(Chat).where(Chat.uid == user_id)
        ).one_or_none()
        return chat_count

    def get_router_user_chats(self, *, router_id: int, user_id: int, db_session: Union[Session, None] = None):
        chats = db_session.exec(
            select(Chat).where((Chat.router_id == router_id) & (Chat.uid == user_id)).order_by(desc(Chat.ctime))
        )
        return chats.all()


class CRUDCustomer(CRUDBase):
    def __init__(self, db_session: Session | None = None) -> None:
        super().__init__(model=Customer, db_session=db_session)

    def get_customer_by_router_id(self, *, router_ids: list[int], db_session: Union[Session, None] = None):
        customers = db_session.exec(
            select(Customer).where(Customer.router_id.in_(router_ids))
        )
        return customers.all()


class CRUDTag(CRUDBase):
    def __init__(self, db_session: Session | None = None) -> None:
        super().__init__(model=Tag, db_session=db_session)



"""Component模块包括组件基类，用户自定义组件需要继承Component类，并至少实现run方法"""

from enum import Enum
from pydantic import BaseModel
from typing import Dict, Any, Optional

from nlpbridge.audio._client import HTTPClient

from nlpbridge.audio.message import Message
from abc import ABC


class Arguments(BaseModel):
    r""""Arguments define Component meta fields"""
    name: str = ""
    tool_desc: Dict[str, Any] = {}

    def extract_values_to_dict(self):
        r"""extract ComponentArguments fields to dict"""
        inputs = {}
        for name, info in self.model_fields.items():
            value = getattr(self, name)
            # 获取 display_name 元数据
            if not info.json_schema_extra:
                continue
            variable_name = info.json_schema_extra.get('variable_name')
            if not variable_name:
                inputs[name] = value
                continue
            # 使用 Enum 成员的实际值
            if isinstance(value, Message):
                inputs[variable_name] = str(value.content)
            elif isinstance(value, Enum):
                inputs[variable_name] = str(value.value)
            else:
                inputs[variable_name] = str(value)
        return inputs


class AudioBase(ABC):
    def __init__(
            self,
            meta: Optional[Arguments] = Arguments(),
            gateway: str = "",
    ):
        self.meta = meta
        self.gateway = gateway
        self._http_client = None

    @property
    def http_client(self):
        if self._http_client is None:
            self._http_client = HTTPClient(self.secret_key, self.gateway)
        return self._http_client




# import BaseLanguageModel from langchain
from langchain_core.callbacks import CallbackManagerForLLMRun
from langchain_core.language_models import BaseLanguageModel
from langchain_core.language_models import BaseLLM

from typing import (
    Any,
    List,
    Optional,
)
from langchain_core.outputs import LLMResult
from langchain_core.runnables import Runnable


class GZUBaseLLM(BaseLLM):

    def invoke(self, input, **kwargs):
        resp = self.generate_prompt(input, **kwargs)
        return resp.llm_output['output']

    def generate_prompt(
            self,
            input: str,
            **kwargs: Any,
        ) -> LLMResult:
        llmresult = LLMResult(
            generations=[],
            llm_output={'output':"Your input is:" + input},
            run=[],
        )
        return llmresult

    def _generate(self, prompts: List[str], stop: List[str] = None, run_manager: CallbackManagerForLLMRun = None, **kwargs: Any) -> LLMResult:
        return prompts[0]
    
    def _llm_type(self) -> str:
        return super()._llm_type()


from typing import(
    List,
    Optional,
)

from zhipuai import ZhipuAI
from langchain_core.embeddings import Embeddings
from langchain_core.pydantic_v1 import (
    BaseModel,
)

class ZhipuAIEmbeddings(BaseModel, Embeddings):
    """ZhipuAI embedding models.

    To use, you should have the
    environment variable ``ZHIPUAI_API_KEY`` set with your API key or pass it
    as a named parameter to the constructor.

    Example:
        .. code-block:: python

            from langchain_zhipuai import ZhipuAIEmbeddings

            embeddding = ZhipuAIEmbeddings()
    """
    model: str = "embedding-2"

    def embed_documents(
        self, texts: List[str], chunk_size: Optional[int] = 0
    ) -> List[List[float]]:
        """Call out to ZhipuAI's embedding endpoint for embedding search docs.

        Args:
            texts: The list of texts to embed.
            chunk_size: The chunk size of embeddings. If None, will use the chunk size
                specified by the class.

        Returns:
            List of embeddings, one for each text.
        """
        embedding_list = []

        # Iterate over each text in the input list
        for text in texts:
            # Call the ZhipuAI API for each individual text
            zhipuai_embedding = ZhipuAI().embeddings.create(model=self.model, input=text)
            
            # Extract the embedding and append it to the embedding_list
            embedding_list.append(zhipuai_embedding.data[0].embedding)
        
        return embedding_list

    async def aembed_documents(
        self, texts: List[str], chunk_size: Optional[int] = 0
    ) -> List[List[float]]:
        """Call out to OpenAI's embedding endpoint async for embedding search docs.

        Args:
            texts: The list of texts to embed.
            chunk_size: The chunk size of embeddings. If None, will use the chunk size
                specified by the class.

        Returns:
            List of embeddings, one for each text.
        """
        zhipuai_embedding = ZhipuAI().embeddings.create(model=self.model,input=texts)
        """
            change embedding_data type of [embedding[object]] to [[float]]
        """
        embedding_list = [item.embedding for item in zhipuai_embedding.data]
        return embedding_list
        

    def embed_query(self, text: str) -> List[float]:
        """Call out to ZhipuAI's embedding endpoint for embedding query text.

        Args:
            text: The text to embed.

        Returns:
            Embedding for the text.
        """
        return self.embed_documents([text])[0]

    async def aembed_query(self, text: str) -> List[float]:
        """Call out to ZhipuAI's embedding endpoint async for embedding query text.

        Args:
            text: The text to embed.

        Returns:
            Embedding for the text.
        """
        embeddings = await self.aembed_documents([text])
        return embeddings[0]